@@ -1,6 +1,7 @@
 /* command.c - Command input and basic commands
  *
  * Copyright (C) 2004, 2005 Oskar Liljeblad
+ * Copyright (C) 2006 Alexey Illarionov <littlesavage@rambler.ru>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -52,6 +53,14 @@
 #include "common/optparser.h"
 #include "microdc.h"
 
+//#define _TRACE
+#if defined(_TRACE)
+#include <stdio.h>
+#define TRACE(x)    printf x; fflush(stdout);
+#else
+#define TRACE(x)
+#endif
+
 static void cmd_retry(int argc, char **argv);
 static void cmd_help(int argc, char **argv);
 static void cmd_exit(int argc, char **argv);
@@ -80,6 +89,8 @@ static void cmd_alias(int argc, char **argv);
 static void cmd_unalias(int argc, char **argv);
 static void cmd_shell(int argc, char **argv);
 static void cmd_lookup(int argc, char **argv);
+static void cmd_share(int argc, char **argv);
+static void cmd_unshare(int argc, char **argv);
 static void shell_command_completion_selector(DCCompletionInfo *ci);
 static void executable_completion_generator(DCCompletionInfo *ci);
 static void alias_completion_generator(DCCompletionInfo *ci);
@@ -304,6 +315,12 @@ command_init(void)
     add_builtin_command("lookup", cmd_lookup, NULL,
         _("lookup HOST ..."),
         _("Lookup the IP address of specified hosts.\n"));
+    add_builtin_command("share", cmd_share, local_path_completion_generator,
+        _("share DIR"),
+        _("Add share directory to the processing list\n"));
+    add_builtin_command("unshare", cmd_unshare, local_path_completion_generator,
+        _("unshare DIR"),
+        _("Remove share directory from the processing list\n"));
 
     add_alias_command("ll", "ls -l");
 }
@@ -532,17 +549,21 @@ executable_completion_generator(DCCompletionInfo *ci)
     char *p;
     const char *file_part;
     char *dir_part;
+    char *conv_word;
     bool path_has_cwd = false;
 
-    get_file_dir_part(ci->word, &dir_part, &file_part);
+    conv_word = main_to_fs_string(ci->word);
+
+    get_file_dir_part(conv_word, &dir_part, &file_part);
 
     path = getenv("PATH");
     if (path == NULL) {
-	/*path = xconfstr(_CS_PATH);
-	if (path == NULL || *path == '\0')*/
-	    return;
+        /*path = xconfstr(_CS_PATH);
+        if (path == NULL || *path == '\0')*/
+        free(conv_word);
+        return;
     } else {
-	path = xstrdup(path);
+        path = xstrdup(path);
     }
 
     for (p1 = path; (p = strsep(&p1, ":")) != NULL; ) {
@@ -554,15 +575,18 @@ executable_completion_generator(DCCompletionInfo *ci)
         if (dh != NULL) {
             struct dirent *ent;
             while ((ent = readdir(dh)) != NULL) { /* ignore errors */
-                if (strleftcmp(ci->word, ent->d_name) == 0) {
+                if (strleftcmp(conv_word, ent->d_name) == 0) {
                     char *full;
                     full = catfiles(p, ent->d_name);
                     if (access(full, X_OK) == 0) { /* ignore errors */
                         struct stat sb;
                         if (stat(full, &sb) >= 0 && S_ISREG(sb.st_mode)) {
                             DCCompletionEntry *ce;
+                            char *conv_fname;
                             
-                            ce = new_completion_entry(ent->d_name, ent->d_name);
+                            conv_fname = fs_to_main_string(ent->d_name);
+                            ce = new_completion_entry(conv_fname, conv_fname);
+			                free(conv_fname);
                             ce->sorting.file_type = DC_TYPE_REG;
                             ptrv_append(ci->results, ce);
                         }
@@ -575,6 +599,8 @@ executable_completion_generator(DCCompletionInfo *ci)
         }
     }
 
+    free(conv_word);
+
     /* local_fs_completion_generator will sort the results. */
     if (*dir_part != '\0' || path_has_cwd) {
         local_fs_completion_generator(ci,DC_CPL_DIR|DC_CPL_EXE|DC_CPL_DOT);
@@ -733,17 +759,47 @@ cmd_status(int argc, char **argv)
         screen_putf(_("Hub users: %s\n"), _("(not logged in)"));
     }
 
+    /*
     screen_putf(_("Pending user connections:\n"));
     hmap_iterator(pending_userinfo, &it);
-    for (c = 0; it.has_next(&it); c++)
-        screen_putf("  %s\n", quotearg(it.next(&it)));
+    for (c = 0; it.has_next(&it); c++) {
+        DCUserInfo* ui = it.next(&it);
+        if (ui != NULL && ui->nick != NULL) {
+            screen_putf("  %s\n", quotearg(ui->nick));
+        } else {
+            screen_putf("  (null)\n");
+        }
+    }
     if (c == 0)
         screen_putf(_("  (none)\n"));
+    */
 
-    screen_putf(_("Share size: %" PRIu64" %s (%s)\n"),
+    c = 0;
+    screen_putf(_("Shared directories:\n"));
+    if (our_filelist != NULL) {
+        hmap_iterator(our_filelist->dir.children, &it);
+        while (it.has_next(&it)) {
+            DCFileList *node = it.next(&it);
+            char* screen_path = fs_to_main_string(node->dir.real_path);
+            screen_putf(_("  %s - %" PRIu64" %s (%s)\n"),
+                        screen_path,
+                        node->size,
+                        ngettext("byte", "bytes", node->size),
+                        human_readable(node->size, sizebuf, human_suppress_point_zero|human_autoscale|human_base_1024|human_SI|human_B, 1, 1));
+            free(screen_path);
+            c++;
+        }
+    }
+    if (c == 0) {
+        screen_putf(_("  (none)\n"));
+    }
+
+    screen_putf(_("Total share size: %" PRIu64" %s (%s)\n"),
                 my_share_size,
                 ngettext("byte", "bytes", my_share_size),
                 human_readable(my_share_size, sizebuf, human_suppress_point_zero|human_autoscale|human_base_1024|human_SI|human_B, 1, 1));
+    screen_putf(_("FileList was updated last time on %s\n"), ctime(&our_filelist_last_update));
+
     screen_putf(_("Bytes received: %" PRIu64 " %s (%s)\n"),
                 bytes_received,
                 ngettext("byte", "bytes", bytes_received),
@@ -752,6 +808,10 @@ cmd_status(int argc, char **argv)
                 bytes_sent,
                 ngettext("byte", "bytes", bytes_sent),
                 human_readable(bytes_sent, sizebuf, human_suppress_point_zero|human_autoscale|human_base_1024|human_SI|human_B, 1, 1));
+
+    if (update_status != NULL) {
+        screen_putf(_("%s\n"), update_status);
+    }
 }
 
 static void
@@ -769,6 +829,9 @@ cmd_say(int argc, char **argv)
 {
     char *t1;
     char *t2;
+    char *hub_my_nick;
+    char *hub_t2;
+    bool utf8 = false;
 
     if (argc <= 1) {
     	screen_putf(_("Usage: %s MESSAGE..\n"), argv[0]);
@@ -779,15 +842,34 @@ cmd_say(int argc, char **argv)
 	return;
     }
     t1 = join_strings(argv+1, argc-1, ' ');
-    if (to_hub_iconv != (iconv_t) -1) {
-        t2 = iconv_alloc(to_hub_iconv, t1);
-        free(t1);
-        t1 = t2;
-    }
+
+    if (t1 == NULL)
+	return;
+
     t2 = escape_message(t1);
     free(t1);
+
+    if (t2 == NULL)
+	return;
+
     /* Don't print on screen - hub will send us the message back. */
-    hub_putf("<%s> %s|", my_nick, t2); /* Ignore error */
+    /*
+    hub_my_nick = main_to_hub_string(my_nick);
+    hub_t2 = main_to_hub_string(t2);
+    */
+    hub_my_nick = try_main_to_hub_string(my_nick);
+    if (hub_my_nick == NULL) {
+        hub_my_nick = main_to_utf8_string(my_nick);
+        utf8 = true;
+    }
+    hub_t2 = try_main_to_hub_string(t2);
+    if (hub_t2 == NULL) {
+        hub_t2 = main_to_utf8_string(t2);
+        utf8 = true;
+    }
+    hub_putf("<%s>%c%s|", hub_my_nick, (utf8 ? '\xA0' : ' '), hub_t2); /* Ignore error */
+    free(hub_my_nick);
+    free(hub_t2);
     free(t2);
 }
 
@@ -814,16 +896,26 @@ cmd_msg(int argc, char **argv)
 
 
     t1 = join_strings(argv+2, argc-2, ' ');
+
+    if ( t1 == NULL)
+	return;
+
     screen_putf("Private to %s: <%s> %s\n", quotearg_n(0, ui->nick), quotearg_n(1, my_nick), t1); /* XXX: quotearg_n(2, t1)? */
-    if (to_hub_iconv != (iconv_t) -1) {
-        t2 = iconv_alloc(to_hub_iconv, t1);
-        free(t1);
-        t1 = t2;
-    }
+
     t2 = escape_message(t1);
     free(t1);
-    hub_putf("$To: %s From: %s $<%s> %s|", ui->nick, my_nick, my_nick, t2); /* Ignore error */
+
+    if (t2 == NULL)
+	return;
+
+    char *hub_my_nick = main_to_hub_string(my_nick);
+    char *hub_to_nick = main_to_hub_string(ui->nick);
+    char *hub_t2 = main_to_hub_string(t2);
     free(t2);
+    hub_putf("$To: %s From: %s $<%s> %s|", hub_to_nick, hub_my_nick, hub_my_nick, hub_t2); /* Ignore error */
+    free(hub_to_nick);
+    free(hub_my_nick);
+    free (hub_t2);
 }
 
 static void
@@ -852,12 +944,12 @@ cmd_connect(int argc, char **argv)
     uint16_t port;
 
     if (argc == 1) {
-    	screen_putf(_("Usage: %s HOST[:PORT]\n"), argv[0]);
-	return;
+        screen_putf(_("Usage: %s HOST[:PORT]\n"), argv[0]);
+        return;
     }
     if (hub_state != DC_HUB_DISCONNECTED) {
-    	screen_putf(_("Connection in progress, disconnect first.\n"));
-	return;
+        screen_putf(_("Connection in progress, disconnect first.\n"));
+        return;
     }
 
     portstr = strchr(argv[1], ':');
@@ -878,10 +970,11 @@ static void
 cmd_disconnect(int argc, char **argv)
 {
     if (hub_state == DC_HUB_DISCONNECTED) {
-    	warn(_("Not connected.\n"));
+        warn(_("Not connected.\n"));
     } else {
-    	warn(_("Disconnecting from hub.\n"));
-	hub_disconnect();
+        warn(_("Disconnecting from hub.\n"));
+        hub_disconnect();
+        hub_set_connected(false);
     }
 }
 
@@ -981,28 +1074,28 @@ static void
 cmd_browse(int argc, char **argv)
 {
     DCUserInfo *ui;
-    char *filename;
+    char *filename = NULL, *xml_filename = NULL, *bzxml_filename = NULL;
     struct stat st;
 
     if (argc == 1) {
-    	if (!browsing_myself && browse_user == NULL) {
-	    screen_putf(_("Not browsing any user.\n"));
-	    return;
-	}
-	browse_none();
-	update_prompt();
-	return;
+        if (!browsing_myself && browse_user == NULL) {
+            screen_putf(_("Not browsing any user.\n"));
+            return;
+        }
+        browse_none();
+        update_prompt();
+        return;
     }
 
     if (strcmp(my_nick, argv[1]) == 0) {
-	browse_none();
-	browse_list = our_filelist;
-	browse_path = xstrdup("/");
-	browse_path_previous = NULL;
-	browse_user = NULL;
-	browsing_myself = true;
-	update_prompt();
-	return;
+        browse_none();
+        browse_list = our_filelist;
+        browse_path = xstrdup("/");
+        browse_path_previous = NULL;
+        browse_user = NULL;
+        browsing_myself = true;
+        update_prompt();
+        return;
     }
 
     ui = hmap_get(hub_users, argv[1]);
@@ -1012,35 +1105,58 @@ cmd_browse(int argc, char **argv)
     }
 
     filename = xasprintf("%s/%s.MyList.DcLst", listing_dir, ui->nick);
-    if (stat(filename, &st) < 0) {
-	if (errno != ENOENT) {
+    xml_filename = xasprintf("%s/%s.files.xml", listing_dir, ui->nick);
+    bzxml_filename = xasprintf("%s/%s.files.xml", listing_dir, ui->nick);
+    if (stat(filename, &st) >= 0) {
+        unlink(filename);
+    }
+    if (stat(filename, &st) >= 0) {
+        unlink(filename);
+    }
+    if (stat(filename, &st) >= 0) {
+        unlink(filename);
+    }
+    if (stat(filename, &st) < 0 &&
+        stat(xml_filename, &st) < 0 &&
+        stat(bzxml_filename, &st) < 0) {
+        if (errno != ENOENT) {
             screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
             free(filename);
+            free(xml_filename);
+            free(bzxml_filename);
             return;
-	}
+        }
 
         free(filename);
-	if (ptrv_find(ui->download_queue, "/MyList.DcLst", (comparison_fn_t) queued_file_cmp) < 0) {
-	    DCQueuedFile *queued = xmalloc(sizeof(DCQueuedFile));
+        free(xml_filename);
+        free(bzxml_filename);
+        if (ptrv_find(ui->download_queue, "/MyList.DcLst", (comparison_fn_t) queued_file_cmp) < 0 &&
+            ptrv_find(ui->download_queue, "/files.xml", (comparison_fn_t) queued_file_cmp) < 0 &&
+            ptrv_find(ui->download_queue, "/files.xml.bz2", (comparison_fn_t) queued_file_cmp) < 0) {
+            DCQueuedFile *queued = xmalloc(sizeof(DCQueuedFile));
 
-	    queued->filename = xstrdup("/MyList.DcLst");
-	    queued->base_path = xstrdup("/");
-	    queued->flag = DC_TF_LIST;
-	    queued->status = DC_QS_QUEUED;
-	    queued->length = UINT64_MAX; /* UINT64_MAX means that the size is unknown */
-	    ptrv_prepend(ui->download_queue, queued);
-	}
-    	if (!has_user_conn(ui, DC_DIR_RECEIVE) && ui->conn_count < DC_USER_MAX_CONN)
-	    hub_connect_user(ui); /* Ignore errors */
-	else
-	    screen_putf(_("No free connections. Queued file for download.\n"));
+            TRACE(("%s:%d: enqueue a new file list to download\n", __FUNCTION__, __LINE__));
 
-	browse_none();
-	browse_user = ui;
-	browsing_myself = false;
-	ui->refcount++;
-	update_prompt();
-	return;
+            queued->filename = xstrdup("/MyList.DcLst"); /* just to have something here */
+            queued->base_path = xstrdup("/");
+            queued->flag = DC_TF_LIST;
+            queued->status = DC_QS_QUEUED;
+            queued->length = UINT64_MAX; /* UINT64_MAX means that the size is unknown */
+            ptrv_prepend(ui->download_queue, queued);
+        } else {
+            TRACE(("%s:%d: there is already filelist in download queue\n", __FUNCTION__, __LINE__));
+        }
+        if (!has_user_conn(ui, DC_DIR_RECEIVE) && ui->conn_count < DC_USER_MAX_CONN)
+            hub_connect_user(ui); /* Ignore errors */
+        else
+            screen_putf(_("No free connections. Queued file for download.\n"));
+
+        browse_none();
+        browse_user = ui;
+        browsing_myself = false;
+        ui->refcount++;
+        update_prompt();
+        return;
     }
 
     browse_none();
@@ -1051,6 +1167,8 @@ cmd_browse(int argc, char **argv)
 
     add_parse_request(browse_list_parsed, filename, xstrdup(ui->nick));
     free(filename);
+    free(xml_filename);
+    free(bzxml_filename);
 }
 
 /* XXX: move queue.c/browse.c? */
@@ -1190,29 +1308,33 @@ cmd_ls(int argc, char **argv)
 {
     OptDetail long_opts[] = {
 	{ "l", OPTP_NO_ARG, 'l' },
+	{ "t", OPTP_NO_ARG, 'l' },
 	{ NULL },
     };
     OptParser *p;
-    bool long_mode = false;
+    int mode = 0;
 
     p = optparser_new(long_opts, -1, 0);
     optparser_parse(p, argc, argv);
     while (optparser_has_next(p)) {
-	switch (optparser_next(p)) {
-	case 'l':
-	    long_mode = true;
-	    break;
-	}
+        switch (optparser_next(p)) {
+        case 'l':
+            mode |= DC_LS_LONG_MODE;
+            break;
+        case 't':
+            mode |= DC_LS_TTH_MODE;
+            break;
+        }
     }
     if (optparser_error(p) != NULL) {
-	screen_putf("%s: %s\n", argv[0], optparser_error(p));
-	optparser_free(p);
-	return;
+        screen_putf("%s: %s\n", argv[0], optparser_error(p));
+        optparser_free(p);
+        return;
     }
     if (browse_list == NULL) {
-	screen_putf(_("Not browsing any user.\n"));
-	optparser_free(p);
-	return;
+        screen_putf(_("Not browsing any user.\n"));
+        optparser_free(p);
+        return;
     }
     if (!optparser_has_next_arg(p)) {
         DCFileList *node;
@@ -1220,15 +1342,15 @@ cmd_ls(int argc, char **argv)
         node = filelist_lookup(browse_list, browse_path);
         /* At the moment, node cannot be NULL here. */
         if (node != NULL)
-            filelist_list(node, long_mode);
+            filelist_list(node, mode);
     }
     while (optparser_has_next_arg(p)) {
-	char *arg = optparser_next_arg(p);
-	bool quoted = false;
-	char *basedir;
-	PtrV *results;
-	DCFileList *basenode;
-	uint32_t d;
+        char *arg = optparser_next_arg(p);
+        bool quoted = false;
+        char *basedir;
+        PtrV *results;
+        DCFileList *basenode;
+        uint32_t d;
 
         results = ptrv_new();
         if (has_leading_slash(arg)) {
@@ -1247,7 +1369,7 @@ cmd_ls(int argc, char **argv)
             fullname = apply_cwd(name);
             node = filelist_lookup(browse_list, fullname);
             if (node != NULL) /* Technically, this shouldn't fail */
-                filelist_list(node, long_mode);
+                filelist_list(node, mode);
             free(fullname);                       
             free(name);
         }
@@ -1469,7 +1591,7 @@ user_info_compare(const void *i1, const void *i2)
     const DCUserInfo *f1 = *(const DCUserInfo **) i1;
     const DCUserInfo *f2 = *(const DCUserInfo **) i2;
 
-    return strcmp(f1->nick, f2->nick);
+    return strcoll(f1->nick, f2->nick);
 }
 
 #define IFNULL(a,x) ((a) == NULL ? (x) : (a))
@@ -1946,3 +2068,73 @@ cmd_lookup(int argc, char **argv)
         add_lookup_request(argv[c], NULL, &hints, lookup_address_looked_up, xstrdup(argv[c]));
     }
 }
+
+static void
+cmd_share(int argc, char **argv)
+{
+    struct stat st;
+    char *dir_fs;
+
+    if (argc > 2) {
+        warn(_("too many arguments\n"));
+        return;
+    }
+
+    if (argc == 1) {
+        screen_putf(_("missing directory argument\n"));
+        return;
+    }
+
+    dir_fs = main_to_fs_string(argv[1]);
+    if (stat(dir_fs, &st) < 0) {
+    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(argv[1]), errstr);
+        free(dir_fs);
+        return;
+    }
+    if (!S_ISDIR(st.st_mode)) {
+    	screen_putf(_("%s: Not a directory\n"), quotearg(argv[1]));
+        free(dir_fs);
+        return;
+    }
+    if (update_request_mq != NULL) {
+        if (!update_request_add_shared_dir(dir_fs)) {
+            screen_putf(_("%s: Cannot process directory - %s\n"), quotearg(argv[1]), errstr);
+        }
+    }
+    free(dir_fs);
+}
+
+static void
+cmd_unshare(int argc, char **argv)
+{
+    struct stat st;
+    char *dir_fs;
+
+    if (argc > 2) {
+        warn(_("too many arguments\n"));
+        return;
+    }
+
+    if (argc == 1) {
+        screen_putf(_("missing directory argument\n"));
+        return;
+    }
+
+    dir_fs = main_to_fs_string(argv[1]);
+    if (stat(dir_fs, &st) < 0) {
+    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(argv[1]), errstr);
+        free(dir_fs);
+        return;
+    }
+    if (!S_ISDIR(st.st_mode)) {
+    	screen_putf(_("%s: Not a directory\n"), quotearg(argv[1]));
+        free(dir_fs);
+        return;
+    }
+    if (update_request_mq != NULL) {
+        if (!update_request_del_shared_dir(dir_fs)) {
+            screen_putf(_("%s: Cannot process directory - %s\n"), quotearg(argv[1]), errstr);
+        }
+    }
+    free(dir_fs);
+}
