@@ -1,6 +1,6 @@
 /* byteq.h - A queue structure holding bytes, designed for I/O tasks
  *
- * Copyright (C) 2004-2005 Oskar Liljeblad
+ * Copyright (C) 2004, 2005 Oskar Liljeblad
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -12,9 +12,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU Library General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #ifndef BYTEQ_H
@@ -30,25 +30,25 @@
 typedef struct _ByteQ ByteQ;
 
 struct _ByteQ {
-    uint8_t *buf;   /* Byte buffer */
-    uint32_t cur;   /* Current number of bytes used in buffer */
-    uint32_t max;   /* Number of bytes (used or not) in buffer */
+    char *buf;	  /* Byte buffer */
+    size_t cur;   /* Current number of bytes used in buffer */
+    size_t max;   /* Number of bytes (used or not) in buffer */
 #ifdef HAVE_FOPENCOOKIE
     FILE *file;     /* Custom stream for speeding up byteq_(v)appendf */
 #endif
 };
 
-ByteQ *byteq_new(uint32_t initial_size);
+ByteQ *byteq_new(size_t initial_size);
 void byteq_free(ByteQ *bq);
-void byteq_append(ByteQ *bq, const uint8_t *data, uint32_t len);
+void byteq_append(ByteQ *bq, void *data, size_t len);
 int byteq_vappendf(ByteQ *bq, const char *format, va_list ap) __attribute__ ((format (printf, 2, 0)));
 int byteq_appendf(ByteQ *bq, const char *format, ...) __attribute__ ((format (printf, 2, 3)));
-void byteq_remove(ByteQ *bq, uint32_t len);
-void byteq_assure(ByteQ *bq, uint32_t max);
+void byteq_remove(ByteQ *bq, size_t len);
+void byteq_assure(ByteQ *bq, size_t max);
 ssize_t byteq_read(ByteQ *bq, int fd);
 ssize_t byteq_full_read(ByteQ *bq, int fd);
-ssize_t byteq_read_upto(ByteQ *bq, int fd, uint32_t len);
-ssize_t byteq_full_read_upto(ByteQ *bq, int fd, uint32_t len);
+ssize_t byteq_read_upto(ByteQ *bq, int fd, size_t len);
+ssize_t byteq_full_read_upto(ByteQ *bq, int fd, size_t len);
 ssize_t byteq_write(ByteQ *bq, int fd);
 ssize_t byteq_full_write(ByteQ *bq, int fd);
 void byteq_clear(ByteQ *bq);
