@@ -100,17 +100,18 @@ ptrv_remove_range(PtrV *pv, uint32_t start, uint32_t end)
     pv->cur -= end - start;
 }
 
-/* Remove the first pointer from the ptrv.
+/* Remove the pointer from the ptrv.
  */
 void *
-ptrv_remove_first(PtrV *pv)
+ptrv_remove(PtrV *pv, uint32_t pos)
 {
-    void *value;
-
-    value = pv->buf[0];
-    pv->cur--;
-    if (pv->cur > 0)
-        memmove(pv->buf, pv->buf+1, pv->cur * sizeof(void *));
+    void *value = NULL;
+    if (pos >= 0 && pos < pv->cur) {
+        value = pv->buf[pos];
+        pv->cur--;
+        if ((pv->cur-pos) > 0)
+            memmove(pv->buf+pos, pv->buf+pos+1, (pv->cur-pos) * sizeof(void *));
+    }
 
     return value;
 }
@@ -120,7 +121,7 @@ ptrv_remove_first(PtrV *pv)
 void
 ptrv_foreach(PtrV *pv, PtrVForeachCallback callback)
 {
-    int c;
+    unsigned c;
 
     for (c = 0; c < pv->cur; c++)
         callback(pv->buf[c]);
