@@ -12,9 +12,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #ifndef QUOTING_H
@@ -29,6 +29,7 @@ char *dequote_words_full(const char *str, bool quoted, bool c_hex_unescape, bool
 int get_word_index(const char *str, int pos);
 char *get_completion_word_dequoted(const char *str, int pos);
 char *get_words_dequoted(const char *str, int index, size_t count);
+int find_completion_word_start(const char *str, int pos); /* = find_word_start_backwards */
 
 const char *find_word_start(const char *str, const char *maxend);
 const char *find_word_end_termchar(const char *str, const char *maxend, char termchar);
@@ -51,4 +52,6 @@ char **get_word_array_dequoted(const char *str, const char *strend, int *argc);
 
 int char_is_quoted(char *string, int index);
 
+int count_unquoted_chars(const char *start, const char *end);
+
 #endif
