@@ -159,15 +159,17 @@ escape_message(const char *str)
 void
 dump_command(const char *header, const char *buf, size_t len)/* XXX: uint32_t=>size_t, same for byteq */
 {
+    const unsigned char *ubuf;
     StrBuf *out;
     uint32_t c;
 
+    ubuf = (const unsigned char *)buf;
     out = strbuf_new();
     for (c = 0; c < len; c++) {
-        if (isprint(buf[c]))
-            strbuf_appendf(out, "%c", buf[c]);
+        if (isprint(ubuf[c]))
+            strbuf_appendf(out, "%c", ubuf[c]);
         else
-            strbuf_appendf(out, "\\x%02x", buf[c] & 0xFF);
+            strbuf_appendf(out, "\\x%02x", ubuf[c] & 0xFF);
     }
     flag_putf(DC_DF_DEBUG, "%s %s\n", header, out->buf);
     strbuf_free(out);
