@@ -1,6 +1,7 @@
 /* fs.c - Local and remote file system management
  *
  * Copyright (C) 2004, 2005 Oskar Liljeblad
+ * Copyright (C) 2006 Alexey Illarionov <littlesavage@rambler.ru>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,6 +20,7 @@
 
 #include <config.h>
 #include <assert.h>		/* ? */
+#include <ctype.h>
 #include <sys/types.h>		/* ? */
 #include <sys/stat.h>		/* ? */
 #include <unistd.h>		/* POSIX */
@@ -46,8 +48,16 @@
 #include "common/strleftcmp.h"
 #include "common/substrcmp.h"
 #include "common/comparison.h"
+#include "iconvme.h"
 #include "microdc.h"
 
+//#define _TRACE
+#if defined(_TRACE)
+#define TRACE(x)    printf x; fflush(stdout);
+#else
+#define TRACE(x)
+#endif
+
 #define MAX_DIR_DEPTH 32
 #define IS_OCT_DIGIT(d) ( (d) >= '0' && (d) <= '7' )
 
@@ -59,7 +69,39 @@ struct _DCFileListIterator {
     uint32_t c;
 };
 
-DCFileList *our_filelist = NULL;
+DCFileList *our_filelist              = NULL;
+time_t      our_filelist_last_update  = 0;
+
+static int
+mkdirs_for_temp_file(char *filename)
+{
+    char *t;
+
+    for (t = filename; *t == '/'; t++);
+    while ((t = strchr(t, '/')) != NULL) {
+	struct stat st;
+
+        *t = '\0';
+	if (stat(filename, &st) < 0) {
+	    if (errno != ENOENT) {
+	    	warn(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
+		return -1;
+	    } else {
+        	if (mkdir(filename, 0777) < 0) {
+    	    	    warn(_("%s: Cannot create directory - %s\n"), quotearg(filename), errstr); 
+		    return -1;
+		} else {
+	    	    if (ptrv_find(delete_dirs, filename, (comparison_fn_t) strcmp) < 0)
+	    		ptrv_append(delete_dirs, xstrdup(filename));
+		}
+	    }
+	}
+        *t = '/';
+	for (; *t == '/'; t++);
+    }
+
+    return 0;
+}
 
 static int
 fs_completion_entry_compare(const void *e1, const void *e2)
@@ -69,7 +111,7 @@ fs_completion_entry_compare(const void *e1, const void *e2)
     if (ce1->sorting.file_type == ce2->sorting.file_type) {
         char *s1 = xasprintf(ce1->display_fmt, ce1->display); /* XXX: this is really slow */
         char *s2 = xasprintf(ce2->display_fmt, ce2->display); /* XXX: this is really slow */
-        int cmp = strcmp(s1, s2);
+        int cmp = strcoll(s1, s2);
         free(s1);
         free(s2);
         return cmp;
@@ -91,16 +133,49 @@ new_file_node(const char *name, DCFileType type, DCFileList *parent)
     node->size = 0;
     switch (node->type) {
     case DC_TYPE_DIR:
+        node->dir.real_path = NULL;
     	node->dir.children = hmap_new();
 	break;
     case DC_TYPE_REG:
-        /* No operation at the moment! */
+        node->reg.has_tth = false;
+        memset(node->reg.tth, 0, sizeof(node->reg.tth));
+        node->reg.mtime = 0;
+        /* No more operation at the moment! */
         break;
     }
 
     return node;
 }
 
+void
+rename_node(DCFileList *node, const char* new_name)
+{
+    char* node_name;
+
+    if (node == NULL || new_name == NULL) {
+        return;
+    }
+
+    node_name = node->name;
+    node->name = xstrdup(new_name);
+
+    if (node->parent != NULL) {
+        hmap_remove(node->parent->dir.children, node_name);
+        hmap_put(node->parent->dir.children, node->name, node);
+    }
+    free(node_name);
+}
+
+void
+set_child_node(DCFileList *parent, DCFileList *child)
+{
+    if (parent == NULL || child == NULL || parent->type != DC_TYPE_DIR) {
+        return;
+    }
+    child->parent = parent;
+    hmap_put(parent->dir.children, child->name, child);
+}
+
 static DCFileList *
 get_child_node(DCFileList *node, const char *path)
 {
@@ -114,16 +189,49 @@ get_child_node(DCFileList *node, const char *path)
 void
 filelist_free(DCFileList *node)
 {
-    switch (node->type) {
-    case DC_TYPE_REG:
-    	break;
-    case DC_TYPE_DIR:
-        hmap_foreach_value(node->dir.children, filelist_free);
-	hmap_free(node->dir.children);
-    	break;
+    if (node != NULL) {
+        switch (node->type) {
+        case DC_TYPE_REG:
+    	    break;
+        case DC_TYPE_DIR:
+            if (node->dir.real_path != NULL)
+                free(node->dir.real_path);
+            hmap_foreach_value(node->dir.children, filelist_free);
+            hmap_free(node->dir.children);
+    	    break;
+        }
+        free(node->name);
+        free(node);
     }
-    free(node->name);
-    free(node);
+}
+
+static DCFileList *
+filelist_lookup_tth(DCFileList *node, const char *tth)
+{
+    if (node->type == DC_TYPE_REG) {
+	unsigned n;
+
+	if (!node->reg.has_tth)
+	    return NULL;
+
+	for (n = 0; n < sizeof(node->reg.tth); n++) {
+	    if ( tolower(node->reg.tth[n]) != tolower(tth[n]))
+		return NULL;
+	}
+	return node;
+    }else {
+	HMapIterator it;
+	DCFileList *res;
+
+	hmap_iterator(node->dir.children, &it);
+	while (it.has_next(&it)) {
+	    DCFileList *subnode = it.next(&it);
+	    res =  filelist_lookup_tth( subnode, tth);
+	    if (res)
+		return res;
+	}
+    }
+    return NULL;
 }
 
 DCFileList *
@@ -133,22 +241,25 @@ filelist_lookup(DCFileList *node, const char *filename)
     char *name;
 
     if (*filename != '/')
-    	return NULL;
+        return NULL;
+
     for (filename++; *filename == '/'; filename++);
+
     end = strchr(filename, '/');
     if (end == NULL) {
-	if (!*filename)
-	    return node;
-    	end = filename + strlen(filename);
+        if (!*filename)
+            return node;
+        end = filename + strlen(filename);
     }
     if (node->type != DC_TYPE_DIR)
-    	return NULL;
+        return NULL;
 
     name = xstrndup(filename, end-filename); /* XXX: strndupa? */
     node = get_child_node(node, name);
     free(name);
+
     if (node != NULL)
-	return (*end == '\0' ? node : filelist_lookup(node, end));
+        return (*end == '\0' ? node : filelist_lookup(node, end));
 
     return NULL;
 }
@@ -195,12 +306,22 @@ filelist_get_path(DCFileList *node)
 static char *
 filelist_get_real_path(DCFileList *node)
 {
-    char *p1;
     char *p2;
 
+    /*
+    char *p1;
     p1 = filelist_get_path(node);
-    p2 = catfiles(share_dir, p1+1); /* p1[0] == '/', we don't want that */
+    p2 = catfiles(share_dir, p1+1); / * p1[0] == '/', we don't want that * /
     free(p1);
+    */
+    if (node->parent != NULL) {
+        /* parent may be only directory */
+
+        assert(node->parent->dir.real_path != NULL);
+        p2 = catfiles(node->parent->dir.real_path, node->name);
+    } else {
+        p2 = xstrdup(node->name);
+    }
     return p2;
 }
 
@@ -218,11 +339,33 @@ file_node_compare(const void *i1, const void *i2)
 {
     const DCFileList *f1 = *(const DCFileList **) i1;
     const DCFileList *f2 = *(const DCFileList **) i2;
+    unsigned char *str1, *str2;
+    unsigned char s1[2], s2[2];
+    int res;
 
     if (f1->type != f2->type)
 	return f1->type - f2->type;
 
-    return strcmp(f1->name, f2->name);
+    str1 = (unsigned char *)f1->name;
+    str2 = (unsigned char *)f2->name;
+
+    s1[1] = s2[1] =  '\0';
+
+    for (; *str1 && *str2; str1++, str2++) {
+	if ((*str1 != *str2)
+		&& (tolower(*str1) != tolower(*str2))) {
+	    s1[0] = tolower(*str1);
+	    s2[0] = tolower(*str2);
+	    res = strcoll(s1, s2);
+	    if (res != 0)
+		return res;
+	}
+    }
+    if ( *str1 )
+	return 1;
+    if ( *str2)
+	return -1;
+    return 0;
 }
 
 static DCFileList **
@@ -252,23 +395,23 @@ void
 filelist_list_recursively(DCFileList *node, char *basepath)
 {
     if (node->type == DC_TYPE_DIR) {
-	DCFileList **items;
-	uint32_t c;
+        DCFileList **items;
+        uint32_t c;
 
         items = get_sorted_file_list(node, NULL);
-	for (c = 0; items[c] != NULL; c++) {
-	    char *path = catfiles(basepath, items[c]->name);
+        for (c = 0; items[c] != NULL; c++) {
+            char *path = catfiles(basepath, items[c]->name);
             filelist_list_recursively(items[c], path);
             free(path);
         }
-	free(items);
+        free(items);
     } else {
-	screen_putf("%7" PRIu64 "M %s\n", (uint64_t) (node->size/(1024*1024)), quotearg(basepath)); /* " */
+        screen_putf("%7" PRIu64 "M %s\n", (uint64_t) (node->size/(1024*1024)), quotearg(basepath)); /* " */
     }
 }
 
 void
-filelist_list(DCFileList *node, bool long_mode)
+filelist_list(DCFileList *node, int mode)
 {
     uint32_t maxlen;
     uint64_t maxsize;
@@ -276,139 +419,180 @@ filelist_list(DCFileList *node, bool long_mode)
     if (node->type == DC_TYPE_DIR) {
         HMapIterator it;
 
-	maxlen = 0;
-	maxsize = 0;
-	hmap_iterator(node->dir.children, &it);
-	while (it.has_next(&it)) {
-    	    DCFileList *subnode = it.next(&it);
+        maxlen = 0;
+        maxsize = 0;
+        hmap_iterator(node->dir.children, &it);
+        while (it.has_next(&it)) {
+            DCFileList *subnode = it.next(&it);
 
-	    switch (subnode->type) {
-	    case DC_TYPE_REG:
-		maxsize = max(maxsize, subnode->size);
-        	maxlen = max(maxlen, strlen(subnode->name));
-        	break;
+            switch (subnode->type) {
+            case DC_TYPE_REG:
+                maxsize = max(maxsize, subnode->size);
+                maxlen = max(maxlen, strlen(subnode->name));
+                break;
             case DC_TYPE_DIR:
-		maxsize = max(maxsize, subnode->size);
-        	maxlen = max(maxlen, strlen(subnode->name)+1);
-        	break;
+                maxsize = max(maxsize, subnode->size);
+                maxlen = max(maxlen, strlen(subnode->name)+1);
+                break;
             }
-	}
+        }
     } else {
-    	maxsize = node->size;
-    	maxlen = strlen(node->name);
+        maxsize = node->size;
+        maxlen = strlen(node->name);
     }
 
-    if (long_mode) {
+    if ((mode & DC_LS_LONG_MODE) != 0) {
         char *format;
 
         format = xasprintf("%%%d" PRIu64 "M %%s%%s\n", ilog10(max(1, maxsize/(1024*1024))));
-	if (node->type == DC_TYPE_DIR) {
-	    DCFileList **items;
-	    uint32_t c;
+        if (node->type == DC_TYPE_DIR) {
+            DCFileList **items;
+            uint32_t c;
 
             items = get_sorted_file_list(node, NULL);
-	    for (c = 0; items[c] != NULL; c++) {
-		switch (items[c]->type) {
-		case DC_TYPE_REG:
-    		    screen_putf(format, (uint64_t) (items[c]->size/(1024*1024)), quotearg(items[c]->name), "");
-		    break;
-		case DC_TYPE_DIR:
-    		    screen_putf(format, (uint64_t) (items[c]->size/(1024*1024)), quotearg(items[c]->name), "/");
-		    break;
-		}
-	    }
-	    free(items);
-	} else {
-	    screen_putf(format, (uint64_t) (node->size/(1024*1024)), quotearg(node->name), "");
-	}
-	free(format);
+            for (c = 0; items[c] != NULL; c++) {
+                switch (items[c]->type) {
+                case DC_TYPE_REG:
+                    screen_putf(format, (uint64_t) (items[c]->size/(1024*1024)), quotearg(items[c]->name), "");
+                    break;
+                case DC_TYPE_DIR:
+                    screen_putf(format, (uint64_t) (items[c]->size/(1024*1024)), quotearg(items[c]->name), "/");
+                    break;
+                }
+	        }
+            free(items);
+        } else {
+	        screen_putf(format, (uint64_t) (node->size/(1024*1024)), quotearg(node->name), "");
+        }
+        free(format);
     } else {
     	if (node->type == DC_TYPE_DIR) {
-	    DCFileList **items;
+            DCFileList **items;
             int cols;
-	    int rows;
-	    int row;
+            int rows;
+            int row;
             int per_row;
             uint32_t count;
 
             items = get_sorted_file_list(node, &count);
             screen_get_size(NULL, &cols);
             per_row = MAX(1, (cols+2)/(maxlen+2));
-	    rows = (count/per_row) + (count%per_row != 0);
+            rows = (count/per_row) + (count%per_row != 0);
 
-	    for (row = 0; row < rows; row++) {
-		uint32_t c;
+            for (row = 0; row < rows; row++) {
+                uint32_t c;
+                for (c = row; c < count; c += rows) {
+                    DCFileList *item = items[c];
+                    int extlen = 0;
+                    int d;
 
-		for (c = row; c < count; c += rows) {
-		    DCFileList *item = items[c];
-		    int extlen = 0;
-		    int d;
-    
-		    switch (item->type) {
-		    case DC_TYPE_REG:
-			extlen = 0;
-			screen_putf("%s", quotearg(item->name));
-			break;
-		    case DC_TYPE_DIR:
-			extlen = 1;
-			screen_putf("%s/", quotearg(item->name));
-			break;
-		    }
-		    if (c+rows < count) {
-			for (d = maxlen-strlen(items[c]->name)-extlen+2; d > 0; d--)
-			    screen_putf(" ");
-		    }
-		}
-		screen_putf("\n");
-	    }
-	    free(items);
-	} else {
-    	    screen_putf("%s\n", quotearg(node->name));
-	}
+                    switch (item->type) {
+                    case DC_TYPE_REG:
+                        extlen = 0;
+                        screen_putf("%s", quotearg(item->name));
+                        break;
+                    case DC_TYPE_DIR:
+                        extlen = 1;
+                        screen_putf("%s/", quotearg(item->name));
+                        break;
+                    }
+                    if (c+rows < count) {
+                    for (d = maxlen-strlen(items[c]->name)-extlen+2; d > 0; d--)
+	                    screen_putf(" ");
+                    }
+                }
+                screen_putf("\n");
+            }
+            free(items);
+        } else {
+            screen_putf("%s\n", quotearg(node->name));
+        }
     }
 }
 
-static void
+void
 dir_to_filelist(DCFileList *parent, const char *path)
 {
     struct dirent *ep;
     DIR *dp;
+    /*
+    char* tth_path = NULL;
+    */
 
     dp = opendir(path);
     if (dp == NULL) {
     	screen_putf(_("%s: Cannot open directory - %s\n"), quotearg(path), errstr);
-	return;
+	    return;
     }
+    
+    parent->dir.real_path = xstrdup(path);
+
+    /*
+    tth_path = catfiles(path, tth_directory_name);
+    */
+
     while ((ep = xreaddir(dp)) != NULL) {
     	struct stat st;
-	char *fullname;
+	    char *fullname;
 
     	if (IS_SPECIAL_DIR(ep->d_name))
-	    continue;
+	        continue;
 
-	/* If we ran into looped symlinked dirs, stat will stop (errno=ELOOP). */
+	    /* If we ran into looped symlinked dirs, stat will stop (errno=ELOOP). */
 
     	fullname = catfiles(path, ep->d_name);
     	if (stat(fullname, &st) < 0) {
-	    screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(fullname), errstr);
+	        screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(fullname), errstr);
+	        free(fullname);
+	        continue;
+	    }
+	    if (S_ISDIR(st.st_mode)) {
+	        DCFileList *node = new_file_node(ep->d_name, DC_TYPE_DIR, parent);
+	        dir_to_filelist(node, fullname);
+	        parent->size += node->size;
+	    }
+	    else if (S_ISREG(st.st_mode)) {
+            /*
+            char *tth_fname = NULL;
+            int tth_fd = -1;
+            */
+	        DCFileList *node = new_file_node(ep->d_name, DC_TYPE_REG, parent);
+	        node->size = st.st_size;
+            node->reg.has_tth = 0;
+            memset(node->reg.tth, 0, sizeof(node->reg.tth));
+            node->reg.mtime = st.st_mtime;
+	        parent->size += node->size;
+
+            /*
+            tth_fname = xasprintf("%s%s%s%s", tth_path, tth_path[0] == '\0' || tth_path[strlen(tth_path)-1] == '/' ? "" : "/", ep->d_name, ".tth");
+            //flag_putf(DC_DF_DEBUG, _("Opening TTH file <%s> for <%s>\n"), tth_fname, ep->d_name);
+            tth_fd = open(tth_fname, O_RDONLY);
+            if (tth_fd >= 0) {
+                uint64_t fsize;
+                time_t mtime, ctime;
+                char tth[39];
+                if (read(tth_fd, &fsize, sizeof(fsize)) == sizeof(fsize) && st.st_size == fsize && 
+                    read(tth_fd, &mtime, sizeof(mtime)) == sizeof(mtime) && st.st_mtime == mtime && 
+                    read(tth_fd, &ctime, sizeof(ctime)) == sizeof(ctime) && st.st_ctime == ctime &&
+                    read(tth_fd, tth, sizeof(tth)) == sizeof(tth)) {
+                    node->reg.has_tth = 1;
+                    memcpy(node->reg.tth, tth, sizeof(node->reg.tth));
+                    //fprintf(stderr, "File <%s> has TTH\n", ep->d_name);
+                }
+                close(tth_fd);
+            }
+            free(tth_fname);
+            */
+	    }
+	    else {
+	        screen_putf(_("%s: Not a regular file or directory, ignoring\n"), quotearg(fullname));
+	    }
 	    free(fullname);
-	    continue;
-	}
-	if (S_ISDIR(st.st_mode)) {
-	    DCFileList *node = new_file_node(ep->d_name, DC_TYPE_DIR, parent);
-	    dir_to_filelist(node, fullname);
-	    parent->size += node->size;
-	}
-	else if (S_ISREG(st.st_mode)) {
-	    DCFileList *node = new_file_node(ep->d_name, DC_TYPE_REG, parent);
-	    node->size = st.st_size;
-	    parent->size += node->size;
-	}
-	else {
-	    screen_putf(_("%s: Not a regular file or directory, ignoring\n"), quotearg(fullname));
-	}
-	free(fullname);
     }
+
+    /*
+    free(tth_path);
+    */
     if (errno != 0)
     	screen_putf(_("%s: Cannot read directory - %s\n"), quotearg(path), errstr);
     if (closedir(dp) < 0)
@@ -418,122 +602,297 @@ dir_to_filelist(DCFileList *parent, const char *path)
 static void
 filelist_to_string(DCFileList *node, StrBuf *sb, int level)
 {
+    char *fname;
+
     if (level != 0)
     	strbuf_append_char_n(sb, level-1, '\t');
 
-    if (node->type == DC_TYPE_REG) {
-   	strbuf_appendf(sb, "%s|%" PRIu64 "\r\n", node->name, node->size); /* " joe sh bug */
-    } else {
-    	HMapIterator it;
+    /* convert filenames from filesystem charset to hub charset */
+    fname = fs_to_hub_string(node->name);
 
-    	if (level != 0)
-   	    strbuf_appendf(sb, "%s\r\n", node->name);
-	hmap_iterator(node->dir.children, &it);
-	while (it.has_next(&it)) {
-	    DCFileList *subnode = it.next(&it);
+    if (node->type == DC_TYPE_REG) {
+        strbuf_appendf(sb, "%s|%" PRIu64 "\r\n", fname, node->size); /* " joe sh bug */
+        free(fname);
+    } else {
+        HMapIterator it;
+
+        if (level != 0)
+            strbuf_appendf(sb, "%s\r\n", fname);
+        free(fname);
+        hmap_iterator(node->dir.children, &it);
+        while (it.has_next(&it)) {
+            DCFileList *subnode = it.next(&it);
             filelist_to_string(subnode, sb, level+1);
-	}
+        }
     }
 }
 
+bool write_filelist_file(DCFileList* root, const char* prefix)
+{
+    StrBuf *sb;
+    char *indata;
+    char *outdata;
+    int fd, xml_fd, bzxml_fd;
+    char *filename;
+    char *xml_filename;
+    char *bzxml_filename;
+    uint32_t len;
+    struct stat st;
+    int i, failed_count;
+    int max_failed_count = 1;
+#if defined(HAVE_LIBXML2)
+    max_failed_count = 3;
+#endif
+
+    if (root != NULL) {
+        sb = strbuf_new();
+	    filelist_to_string(root, sb, 0);
+        len = strbuf_length(sb);
+        indata = strbuf_free_to_string(sb);
+        outdata = huffman_encode((uint8_t *) indata, len, &len);
+        free(indata);
+
+        filename = xasprintf("%s%s%sMyList.DcLst", listing_dir, listing_dir[0] == '\0' || listing_dir[strlen(listing_dir)-1] == '/' ? "" : "/", prefix == NULL ? "" : prefix);
+        mkdirs_for_temp_file(filename); /* Ignore errors */
+
+        if (stat(filename, &st) < 0) {
+    	    if (errno != ENOENT)
+	        warn(_("%s: Cannot get file status - %s\n"), filename, errstr);
+        } else {
+	        if (unlink(filename) < 0)
+	            warn(_("%s: Cannot remove file - %s\n"), filename, errstr);
+        }
+        i = ptrv_find(delete_files, filename, (comparison_fn_t) strcmp);
+        if (i >= 0)
+    	    ptrv_remove_range(delete_files, i, i+1);
+
+#if defined(HAVE_LIBXML2)
+        xml_filename = xasprintf("%s%s%sfiles.xml", listing_dir, listing_dir[0] == '\0' || listing_dir[strlen(listing_dir)-1] == '/' ? "" : "/", prefix == NULL ? "" : prefix);
+        mkdirs_for_temp_file(xml_filename); /* Ignore errors */
+        if (stat(xml_filename, &st) < 0) {
+    	    if (errno != ENOENT)
+	        warn(_("%s: Cannot get file status - %s\n"), xml_filename, errstr);
+        } else {
+	        if (unlink(xml_filename) < 0)
+	            warn(_("%s: Cannot remove file - %s\n"), xml_filename, errstr);
+        }
+        i = ptrv_find(delete_files, xml_filename, (comparison_fn_t) strcmp);
+        if (i >= 0)
+    	    ptrv_remove_range(delete_files, i, i+1);
+
+
+        bzxml_filename = xasprintf("%s%s%sfiles.xml.bz2", listing_dir, listing_dir[0] == '\0' || listing_dir[strlen(listing_dir)-1] == '/' ? "" : "/", prefix == NULL ? "" : prefix);
+        mkdirs_for_temp_file(bzxml_filename); /* Ignore errors */
+        if (stat(bzxml_filename, &st) < 0) {
+    	    if (errno != ENOENT)
+	        warn(_("%s: Cannot get file status - %s\n"), bzxml_filename, errstr);
+        } else {
+	        if (unlink(bzxml_filename) < 0)
+	            warn(_("%s: Cannot remove file - %s\n"), bzxml_filename, errstr);
+        }
+        i = ptrv_find(delete_files, bzxml_filename, (comparison_fn_t) strcmp);
+        if (i >= 0)
+    	    ptrv_remove_range(delete_files, i, i+1);
+#endif
+
+        failed_count = 0;
+        fd = open(filename, O_CREAT|O_EXCL|O_WRONLY, 0666);
+        if (fd < 0) {
+    	    screen_putf(_("%s: Cannot open file for writing - %s\n"), quotearg(filename), errstr);
+            failed_count++;
+        }
+#if defined(HAVE_LIBXML2)
+        xml_fd = open(xml_filename, O_CREAT|O_EXCL|O_WRONLY, 0666);
+        if (xml_fd < 0) {
+    	    screen_putf(_("%s: Cannot open file for writing - %s\n"), quotearg(xml_filename), errstr);
+            failed_count++;
+        }
+        bzxml_fd = open(bzxml_filename, O_CREAT|O_EXCL|O_WRONLY, 0666);
+        if (bzxml_fd < 0) {
+    	    screen_putf(_("%s: Cannot open file for writing - %s\n"), quotearg(bzxml_filename), errstr);
+            failed_count++;
+        }
+#endif
+        if (failed_count == max_failed_count) {
+	        filelist_free(root);
+	        free(outdata);
+	        free(filename);
+#if defined(HAVE_LIBXML2)
+	        free(xml_filename);
+	        free(bzxml_filename);
+#endif
+	        return false;
+        }
+        /*
+        if (ptrv_find(delete_files, filename, (comparison_fn_t) strcmp) < 0)
+    	    ptrv_append(delete_files, xstrdup(filename));
+#if defined(HAVE_LIBXML2)
+        if (ptrv_find(delete_files, xml_filename, (comparison_fn_t) strcmp) < 0)
+    	    ptrv_append(delete_files, xstrdup(xml_filename));
+        if (ptrv_find(delete_files, bzxml_filename, (comparison_fn_t) strcmp) < 0)
+    	    ptrv_append(delete_files, xstrdup(bzxml_filename));
+#endif
+        */
+
+        failed_count = 0;
+        if (fd < 0 || full_write(fd, outdata, len) < len) {
+            if (fd >= 0)
+    	        screen_putf(_("%s: Cannot write to file - %s\n"), quotearg(filename), errstr);
+            failed_count++;
+        }
+        free(outdata);
+
+#if defined(HAVE_LIBXML2)
+        if (xml_fd < 0 || write_xml_filelist(xml_fd, root) < 0) {
+            if (xml_fd >= 0) {
+    	        screen_putf(_("%s: Cannot write to file - %s\n"), quotearg(xml_filename), errstr);
+            }
+            failed_count++;
+        }
+        if (bzxml_fd < 0 || write_bzxml_filelist(bzxml_fd, root) < 0) {
+            if (bzxml_fd >= 0) {
+    	        screen_putf(_("%s: Cannot write to file - %s\n"), quotearg(bzxml_filename), errstr);
+            }
+            failed_count++;
+        }
+#endif
+        if (failed_count == max_failed_count) {
+	        filelist_free(root); 
+	        free(filename);
+#if defined(HAVE_LIBXML2)
+	        free(xml_filename);
+	        free(bzxml_filename);
+#endif
+	        return false;
+        }
+
+        if (close(fd) < 0)
+    	    screen_putf(_("%s: Cannot close file - %s\n"), quotearg(filename), errstr);
+        free(filename);
+
+#if defined(HAVE_LIBXML2)
+        if (close(xml_fd) < 0)
+    	    screen_putf(_("%s: Cannot close file - %s\n"), quotearg(xml_filename), errstr);
+        free(xml_filename);
+
+        // we don't need to close this file - it was closed by write_bzxml_filelist() function.
+        free(bzxml_filename);
+#endif
+
+        return true;
+    }
+    return false;
+}
+
 bool
 filelist_create(const char *basedir)
 {
-    StrBuf *sb;
-    char *indata;
-    char *outdata;
-    int fd;
-    char *filename;
-    uint32_t len;
-    struct stat st;
-    int i;
-    DCFileList *root;
+    char* conv_basedir;
+    DCFileList *root = new_file_node("", DC_TYPE_DIR, NULL);
 
-    root = new_file_node("", DC_TYPE_DIR, NULL);
-
-    sb = strbuf_new();
     if (basedir != NULL) {
     	screen_putf(_("Scanning directory %s\n"), quotearg(basedir));
+        //screen_sync();
     	dir_to_filelist(root, basedir);
-	filelist_to_string(root, sb, 0);
     }
+	/*  basedir for dir_to_filelist must be in filesystem encoding */
+	conv_basedir = main_to_fs_string(basedir);
+	dir_to_filelist(root, conv_basedir);
+    free(conv_basedir);
 
-    len = strbuf_length(sb);
-    indata = strbuf_free_to_string(sb);
-    outdata = huffman_encode((uint8_t *) indata, len, &len);
-    free(indata);
-
-    filename = catfiles(listing_dir, "MyList.DcLst");
-    mkdirs_for_file(filename, true); /* Ignore errors */
-    if (stat(filename, &st) < 0) {
-    	if (errno != ENOENT)
-	    warn(_("%s: Cannot get file status - %s\n"), filename, errstr);
-    } else {
-	if (unlink(filename) < 0)
-	    warn(_("%s: Cannot remove file - %s\n"), filename, errstr);
-    }
-    i = ptrv_find(delete_files, filename, (comparison_fn_t) strcmp);
-    if (i >= 0)
-    	ptrv_remove_range(delete_files, i, i+1);
-
-    fd = open(filename, O_CREAT|O_EXCL|O_WRONLY, 0666);
-    if (fd < 0) {
-    	screen_putf(_("%s: Cannot open file for writing - %s\n"), quotearg(filename), errstr);
-	filelist_free(root);
-	free(outdata);
-	free(filename);
-	return false;
-    }
-    if (ptrv_find(delete_files, filename, (comparison_fn_t) strcmp) < 0)
-    	ptrv_append(delete_files, xstrdup(filename));
-
-    if (full_write(fd, outdata, len) < len) {
-    	screen_putf(_("%s: Cannot write to file - %s\n"), quotearg(filename), errstr);
-	filelist_free(root); 
-	free(outdata);
-	free(filename);
-	return false;
-    }
-    if (close(fd) < 0)
-    	screen_putf(_("%s: Cannot close file - %s\n"), quotearg(filename), errstr);
-    free(filename);
-    free(outdata);
 
     if (our_filelist != NULL)
     	filelist_free(our_filelist);
     our_filelist = root;
     my_share_size = our_filelist->size;
 
-    return true;
+    return write_filelist_file(root, NULL);
 }
 
 /* Find the physical path of a file that the remote end
  * wants to download from us.
  */
 char *
-resolve_upload_file(DCUserInfo *ui, const char *name)
+resolve_upload_file(DCUserInfo *ui, DCAdcgetType ul_type, const char *name, DCTransferFlag* flag, uint64_t* size)
 {
     DCFileList *node;
 
-    /* Note: 'name' will have been translated to local slashes
-     * by a call to translate_remote_to_local prior to calling
-     * this function.
-     */
+    if (ul_type == DC_ADCGET_FILE) {
+        /* Note: 'name' will have been translated to local slashes
+         * by a call to translate_remote_to_local prior to calling
+         * this function.
+         */
 
-    /* Skip leading slashes, all but one */
-    if (name[0] != '/')
-	return NULL;
-    for (; name[1] == '/'; name++);
+        /* Skip leading slashes, all but one */
+        if (name[0] != '/')
+            return NULL;
+        for (; name[1] == '/'; name++);
 
-    if (strcmp(name, "/MyList.DcLst") == 0)
-        return catfiles(listing_dir, name);
+        if (strcmp(name, "/MyList.DcLst") == 0
+#if defined(HAVE_LIBXML2)
+            || strcmp(name, "/files.xml") == 0 || strcmp(name, "/files.xml.bz2") == 0
+#endif
+            ) {
+            char* filename = catfiles(listing_dir, name);
+            if (flag != NULL) {
+                *flag = DC_TF_LIST;
+            }
+            if (size != NULL) {
+                struct stat st;
+                if (stat(filename, &st) == 0) {
+                    *size = st.st_size;
+                } else {
+                    *size = 0;
+                }
+            }
+            return filename;
+        }
+    }
 
-    if (our_filelist == NULL || share_dir == NULL)
+    if (our_filelist == NULL)
         return NULL;
 
-    node = filelist_lookup(our_filelist, name);
+    if (ul_type == DC_ADCGET_FILE) 
+        node = filelist_lookup(our_filelist, name);
+    else
+        node = filelist_lookup_tth(our_filelist, name);
+
     if (node == NULL)
         return NULL;
+    if (ul_type == DC_ADCGET_TTHL) {
+        return NULL;
+/*
+	StrBuf *sb;
+
+	sb = strbuf_new();
+
+	strbuf_append_char(sb, '/');
+	strbuf_append(sb, tth_directory_name);
+	if ( tth_directory_name[strlen(tth_directory_name) - 1] != '/' )
+	    strbuf_append_char(sb, '/');
+	strbuf_append(sb, node->name);
+	strbuf_append(sb, ".tth");
+
+	node = node->parent;
+	while (node != NULL) {
+	    strbuf_prepend(sb, node->name);
+	    strbuf_prepend_char(sb, '/');
+	    node = node->parent;
+	}
+
+	strbuf_prepend(sb, share_dir);
+	
+	return strbuf_free_to_string(sb);
+*/
+
+    } else {
+        if (flag != NULL) {
+            *flag = DC_TF_NORMAL;
+        }
+        if (size != NULL) {
+            *size = node->size;
+        }
+    }
     return filelist_get_real_path(node);
 }
 
@@ -541,19 +900,29 @@ char *
 resolve_download_file(DCUserInfo *ui, DCQueuedFile *queued)
 {
     char *filename;
-    char *tmp;
+    char *tmp, *tmp2;
 
     if (queued->flag == DC_TF_LIST) {
-        tmp = xasprintf("%s.%s", ui->nick, base_name(queued->filename));
-        filename = catfiles(listing_dir, tmp);
+        tmp = xasprintf("%s", ui->nick);
+        tmp2 = catfiles(listing_dir, tmp);
         free(tmp);
-	ptrv_append(delete_files, xstrdup(filename));
-        mkdirs_for_file(filename, true); /* Ignore errors */
+
+        filename = main_to_fs_string(tmp2);
+        free(tmp2);
+
+        if (filename) {
+            mkdirs_for_temp_file(filename); /* Ignore errors */
+        }
     } else {
         tmp = catfiles(download_dir, queued->filename + strlen(queued->base_path));
-        filename = xasprintf("%s.part", tmp);
+        tmp2 = xasprintf("%s.part", tmp);
         free(tmp);
-        mkdirs_for_file(filename, false); /* Ignore errors */
+
+        filename = main_to_fs_string(tmp2);
+	    free(tmp2);
+
+        if (filename)
+            mkdirs_for_file(filename); /* Ignore errors */
     }
 
     return filename;
@@ -585,7 +954,7 @@ translate_local_to_remote(const char *localname)
     localname++;
     for (; *localname != '\0'; localname++) {
     	if (*localname == '/')
-	    strbuf_append_char(sb, '\\');
+            strbuf_append_char(sb, '\\');
 	else
 	    strbuf_append_char(sb, *localname);
     }
@@ -812,8 +1181,11 @@ static void
 add_local_wildcard_result(char *name, DCFSCompletionFlags flags, DCCompletionInfo *ci)
 {
     struct stat st;
+    char *name_fs;
 
-    if (lstat(name, &st) == 0) {
+    name_fs = main_to_fs_string(name);
+
+    if (lstat(name_fs, &st) == 0) {
         DCCompletionEntry *entry = NULL;
         char *input;
         bool dquoted = (ci->word_full[0] == '"');
@@ -825,7 +1197,7 @@ add_local_wildcard_result(char *name, DCFSCompletionFlags flags, DCCompletionInf
                 entry->sorting.file_type = DC_TYPE_DIR;
             }
         } else if (S_ISLNK(st.st_mode)) {
-            if (stat(name, &st) == 0 && S_ISDIR(st.st_mode)) {
+            if (stat(name_fs, &st) == 0 && S_ISDIR(st.st_mode)) {
                 if (flags & DC_CPL_DIR) {
                     entry = new_completion_entry_full(input, name, "%s", "%s@", false, true);
                     if (dquoted)
@@ -844,7 +1216,7 @@ add_local_wildcard_result(char *name, DCFSCompletionFlags flags, DCCompletionInf
             }
         } else if (S_ISREG(st.st_mode)) {
             if ((flags & DC_CPL_REG) || (flags & DC_CPL_EXE)) {
-                if ((access(name, X_OK) == 0) == ((flags & DC_CPL_EXE) != 0)) {
+                if ((access(name_fs, X_OK) == 0) == ((flags & DC_CPL_EXE) != 0)) {
                     entry = new_completion_entry_full(input, name, "%s", "%s", true, true);
                     entry->sorting.file_type = DC_TYPE_REG;
                 } else if ((flags & DC_CPL_EXE) == 0) {
@@ -853,6 +1225,7 @@ add_local_wildcard_result(char *name, DCFSCompletionFlags flags, DCCompletionInf
                 }
             }
         }
+        free(name_fs);
 
         if (entry != NULL) {
             ptrv_append(ci->results, entry);
@@ -1001,58 +1374,76 @@ local_wildcard_complete(char *matchpath, bool *quotedptr, char *basedir, DCFSCom
 {
     char *matchcomp;
     struct stat sb;
-    char *fullpath;
+    char *fullpath_fs, *fullpath;
     DIR *dh;
     struct dirent *de;
+    char *basedir_fs;
+    char *fname_loc;
+
+    basedir_fs = main_to_fs_string(basedir);
 
     if (dircomp_to_fnmatch_str(&matchpath, quotedptr, &matchcomp)) {
-        dh = opendir(*basedir == '\0' ? "." : basedir);
+        dh = opendir(*basedir_fs == '\0' ? "." : basedir_fs);
         if (dh != NULL) {
             while ((de = readdir(dh)) != NULL) {
-                if (fnmatch(matchcomp, de->d_name, FNM_PERIOD) == 0) {
+                /* convert filename from filesystem charset for fnmatch*/
+                fname_loc = fs_to_main_string(de->d_name);
+
+                if (fnmatch(matchcomp, fname_loc, FNM_PERIOD) == 0) {
                     if (*matchpath != '\0') {
-                        fullpath = concat_filenames(basedir, de->d_name);
-                        if (stat(fullpath, &sb) == 0 && S_ISDIR(sb.st_mode))
+                        fullpath_fs = concat_filenames(basedir_fs, de->d_name);
+                        fullpath = concat_filenames(basedir, fname_loc);
+                        if (stat(fullpath_fs, &sb) == 0 && S_ISDIR(sb.st_mode))
                             local_wildcard_complete(matchpath, quotedptr, fullpath, flags, ci, true);
+                        free(fullpath_fs);
                         free(fullpath);
                     } else {
-                        add_local_wildcard_result(concat_filenames(basedir, de->d_name), flags, ci);
+                        add_local_wildcard_result(concat_filenames(basedir, fname_loc), flags, ci);
                     }
                 }
+
+                free(fname_loc);
             }
             closedir(dh);
         }
     } else {
         fullpath = concat_filenames(basedir, matchcomp);
+        fullpath_fs = main_to_fs_string(fullpath);
         if (*matchpath != '\0') {
-            if (lstat(fullpath, &sb) == 0)
+            if (lstat(fullpath_fs, &sb) == 0)
                 local_wildcard_complete(matchpath, quotedptr, fullpath, flags, ci, found_wc);
         } else {
             /* If the string we are completing had wild cards, attempt to expand the string
              * first rather than generating multiple possible completion results.
              * This behavior is similar to that of GNU readline.
              */
-	    if (found_wc) {
-	        if (*matchcomp == '\0') { /* completion word ends in slash */
-	            add_local_wildcard_result(concat_filenames(basedir, matchcomp), flags, ci);
-                } else if (/*strcmp(basedir, "/") != 0 &&*/ lstat(fullpath, &sb) == 0) {
+            if (found_wc) {
+	            if (*matchcomp == '\0') { /* completion word ends in slash */
+	                add_local_wildcard_result(concat_filenames(basedir, matchcomp), flags, ci);
+                } else if (/*strcmp(basedir, "/") != 0 &&*/ lstat(fullpath_fs ? fullpath_fs : fullpath, &sb) == 0) {
                     add_local_wildcard_result(concat_filenames(basedir, matchcomp), flags, ci);
                 }
             } else {
-                dh = opendir(*basedir == '\0' ? "." : basedir);
+                dh = opendir(*basedir_fs == '\0' ? "." : basedir_fs);
                 if (dh != NULL) {
                     while ((de = readdir(dh)) != NULL) {
                         if ((de->d_name[0] == '.') != (matchcomp[0] == '.'))
                             continue;
-                        if (strleftcmp(matchcomp, de->d_name) == 0)
-                            add_local_wildcard_result(concat_filenames(basedir, de->d_name), flags, ci);
+
+                        fname_loc = fs_to_main_string(de->d_name);
+
+                        if (strleftcmp(matchcomp, fname_loc ? fname_loc : de->d_name) == 0)
+                            add_local_wildcard_result(concat_filenames(basedir, fname_loc), flags, ci);
+                        free(fname_loc);
                     }
                     closedir(dh);
                 }
             }
         }
         free(fullpath);
+        free(fullpath_fs);
     }
+    free(basedir_fs);
     free(matchcomp);
 }
 
