@@ -148,6 +148,7 @@ huffman_encode(const uint8_t *data, uint32_t data_size, uint32_t *out_size)
     uint64_t bits;
     uint64_t keybits;
     uint64_t bit_pos;
+    uint8_t *bitdata;
 
     if (data_size == 0) {
     	char *outdata = xmemdup("HE3\xD\0\0\0\0\0\0\0", 11);
@@ -215,8 +216,9 @@ huffman_encode(const uint8_t *data, uint32_t data_size, uint32_t *out_size)
     }
 
     bits = BYTE_BOUNDARY(bits) + BYTE_BOUNDARY(keybits);
-    uint8_t bitdata[bits/8];
-    memset(bitdata, 0, bits/8);
+    bitdata = (uint8_t *)calloc(1, bits/8);
+    if (!bitdata)
+	    return NULL;
     bit_pos = 0;
     for (c = 0; c < 256; c++) {
     	if (counts[c] != 0)
@@ -228,9 +230,10 @@ huffman_encode(const uint8_t *data, uint32_t data_size, uint32_t *out_size)
 	add_bits(bitdata, &bit_pos, bitnodes[ch].data, bitnodes[ch].bitcount);
     }
 
-    strbuf_append_data(out, &bitdata, bits/8);
+    strbuf_append_data(out, bitdata, bits/8);
     free_encode_node(rootnode);
     *out_size = strbuf_length(out);
+    free(bitdata);
     return strbuf_free_to_string(out);
 }
 
