@@ -72,6 +72,7 @@ bool browsing_myself = false;
 /* misc */
 uint64_t bytes_received = 0;
 uint64_t bytes_sent = 0;
+uint64_t max_speed = 0, prev_max_speed = 0;
 uint16_t listen_port = 0;
 bool running = true;
 fd_set read_fds;
@@ -86,8 +87,10 @@ HMap *pending_userinfo = NULL;	/* users we want to communicate with, either
                                  * us connecting to them or they connecting
                                  * to us. (UserInfo->nick => UserInfo) */
 
+
 /* XXX: change to ("Copyright %s %d Oskar Liljeblad", gettext("(C)"), 2005) */
-const char version_etc_copyright[] = "Copyright (C) 2004, 2005 Oskar Liljeblad";
+const char version_etc_copyright[] =
+    "Copyright (C) 2006 Vladimir Chugunov, based on Oskar Liljeblad's microdc 0.11.0\nmicrodc is copyright (C) 2004, 2005 Oskar Liljeblad.";
 
 static uint32_t user_conn_unknown_last = 0;
 static PtrV *user_conn_unknown_free = NULL;
@@ -126,18 +129,18 @@ warn_socket_error(int res, bool write, const char *format, ...)
 {
     char *subject;
     va_list args;
-    
+
     va_start(args, format);
     subject = xvasprintf(format, args);
     va_end(args);
     if (write) {
-	warn("Cannot send to %s - %s\n", subject, errstr);
+        warn("Cannot send to %s - %s\n", subject, errstr);
     } else {
-	if (res < 0) {
-	    warn("Cannot receive from %s - %s\n", subject, errstr);
-	} else {
-	    warn("Disconnected from %s.\n", subject);
-	}
+        if (res < 0) {
+            warn("Cannot receive from %s - %s\n", subject, errstr);
+        } else {
+            warn("Disconnected from %s.\n", subject);
+        }
     }
     free(subject);
 }
@@ -293,14 +296,31 @@ display_transfer_ended_msg(bool upload, DCUserConn *uc, bool success, const char
 static void
 handle_ended_upload(DCUserConn *uc, bool success, const char *reason)
 {
-    bytes_sent += uc->transfer_pos - uc->transfer_start;
-    if (uc->occupied_slot) {
+    time_t now = (time_t)-1;
+    uint64_t len = uc->transfer_pos - uc->transfer_start;
+    uint64_t speed = len/MAX(now - uc->transfer_time, 1);
+
+    time(&now);
+    if (speed > max_speed) {
+        prev_max_speed = max_speed;
+        max_speed = speed;
+    }
+
+    bytes_sent += len;
+
+    if (uc->occupied_slot || uc->occupied_minislot) {
         /* Only display "Upload failed" if "Starting upload" was
          * displayed previously.
          */
-	display_transfer_ended_msg(true, uc, success, " (%s)", reason);
-        used_ul_slots--;
-        uc->occupied_slot = false;
+	    display_transfer_ended_msg(true, uc, success, " (%s)", reason);
+        if (uc->occupied_slot) {
+            used_ul_slots--;
+            uc->occupied_slot = false;
+        }
+        if (uc->occupied_minislot) {
+            used_mini_slots--;
+            uc->occupied_minislot = false;
+        }
     }
     free(uc->transfer_file);
     uc->transfer_file = NULL;
@@ -322,40 +342,42 @@ handle_ended_download(DCUserConn *uc, bool success, const char *reason)
     if (uc->queued_valid) {
         DCQueuedFile *queued;
 
+        /*
         queued = uc->info->download_queue->buf[uc->queue_pos];
         uc->queue_pos++;
+        */
+        queued = ptrv_remove(uc->info->download_queue, uc->queue_pos);
         uc->queued_valid = false;
         if (success) {
             queued->status = DC_QS_DONE;
             if (queued->flag == DC_TF_LIST) {
+                if (uc->local_file != NULL) {
+                    ptrv_append(delete_files, xstrdup(uc->local_file));
+                }
                 if (browse_user != NULL && strcmp(browse_user->nick, uc->info->nick) == 0 && browse_list == NULL) {
-                    char *filename;
-
-                    filename = xasprintf("%s/%s.MyList.DcLst", listing_dir, browse_user->nick);
-		    add_parse_request(browse_list_parsed, filename, xstrdup(browse_user->nick));
-		    free(filename);
+                    add_parse_request(browse_list_parsed, uc->local_file, xstrdup(browse_user->nick));
                 }
             } else {
-                char *local_file = resolve_download_file(uc->info, queued);
-                char *final_file = xstrndup(local_file, strlen(local_file)-5);
-                if (safe_rename(local_file, final_file) != 0) {
-                    warn(_("%s: Cannot rename file to %s - %s\n"), quotearg_n(0, local_file), quote_n(1, final_file), errstr);
+                char *final_file = xstrndup(uc->local_file, strlen(uc->local_file)-5);
+                if (safe_rename(uc->local_file, final_file) != 0) {
+                    warn(_("%s: Cannot rename file to %s - %s\n"), quotearg_n(0, uc->local_file), quote_n(1, final_file), errstr);
                     reason = _("cannot rename file"); /* XXX: would like a more elaborate error message here perhaps? */
                     queued->status = DC_QS_ERROR;
                     success = false;
                 }
                 free(final_file);
-                free(local_file);
             }
         } else {
             queued->status = DC_QS_ERROR;
         }
-	display_transfer_ended_msg(false, uc, success, " (%s)", reason);
+        display_transfer_ended_msg(false, uc, success, " (%s)", reason);
     } else {
-	display_transfer_ended_msg(false, uc, success, " but unqueued (%s)", reason);
+        display_transfer_ended_msg(false, uc, success, " but unqueued (%s)", reason);
     }
     free(uc->transfer_file);
+    free(uc->local_file);
     uc->transfer_file = NULL;
+    uc->local_file = NULL;
     uc->transfer_start = 0;
     uc->transfer_pos = 0;
     uc->transferring = false;
@@ -395,8 +417,10 @@ user_connection_new(struct sockaddr_in *addr, int user_socket)
     uc->pid = pid;
     uc->info = NULL;
     uc->occupied_slot = 0;
+    uc->occupied_minislot = 0;
     uc->dir = DC_DIR_UNKNOWN;
     uc->transfer_file = NULL;
+    uc->local_file = NULL;
     uc->transfer_start = 0;
     uc->transfer_pos = 0;
     uc->transfer_total = 0;
@@ -537,11 +561,11 @@ user_result_fd_readable(DCUserConn *uc)
         return;
     }
     while (msgq_has_complete_msg(uc->get_mq)) {
-	int id;
+        int id;
 
         msgq_peek(uc->get_mq, MSGQ_INT, &id, MSGQ_END);
-	switch (id) {
-	case DC_MSG_SCREEN_PUT: {
+        switch (id) {
+        case DC_MSG_SCREEN_PUT: {
             char *msg;
             uint32_t flag;
 
@@ -604,30 +628,36 @@ user_result_fd_readable(DCUserConn *uc)
         case DC_MSG_TRANSFER_STATUS:
     	    msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_INT64, &uc->transfer_pos, MSGQ_END);
     	    break;
-        case DC_MSG_TRANSFER_START:
-    	    msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_INT64, &uc->transfer_start, MSGQ_INT64, &uc->transfer_total, MSGQ_END);
+        case DC_MSG_TRANSFER_START: {
+            char *share_filename, *local_filename;
+    	    msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_STR, &local_filename, MSGQ_STR, &share_filename, MSGQ_INT64, &uc->transfer_start, MSGQ_INT64, &uc->transfer_total, MSGQ_END);
     	    if (uc->transfer_start != 0) {
-	        flag_putf(DC_DF_CONNECTIONS, _("%s: Starting %s of %s (%" PRIu64 " of %" PRIu64 " %s).\n"),
-                        quotearg_n(0, uc->info->nick),
-                        uc->dir == DC_DIR_SEND ? _("upload") : _("download"),
-                        quote_n(1, base_name(uc->transfer_file)),
-                        uc->transfer_total - uc->transfer_start,
-                        uc->transfer_total,
-                        ngettext("byte", "bytes", uc->transfer_total));
+                flag_putf(DC_DF_CONNECTIONS, _("%s: Starting %s of %s (%" PRIu64 " of %" PRIu64 " %s).\n"),
+                          quotearg_n(0, uc->info->nick),
+                          uc->dir == DC_DIR_SEND ? _("upload") : _("download"),
+                          quote_n(1, base_name(share_filename)),
+                          uc->transfer_total - uc->transfer_start,
+                          uc->transfer_total,
+                          ngettext("byte", "bytes", uc->transfer_total));
             } else {
-	        flag_putf(DC_DF_CONNECTIONS, _("%s: Starting %s of %s (%" PRIu64 " %s).\n"),
-                        quotearg_n(0, uc->info->nick),
-                        uc->dir == DC_DIR_SEND ? _("upload") : _("download"),
-                        quote_n(1, base_name(uc->transfer_file)),
-                        uc->transfer_total,
-                        ngettext("byte", "bytes", uc->transfer_total));
+                flag_putf(DC_DF_CONNECTIONS, _("%s: Starting %s of %s (%" PRIu64 " %s).\n"),
+                          quotearg_n(0, uc->info->nick),
+                          uc->dir == DC_DIR_SEND ? _("upload") : _("download"),
+                          quote_n(1, base_name(share_filename)),
+                          uc->transfer_total,
+                          ngettext("byte", "bytes", uc->transfer_total));
             }
+            free(uc->transfer_file);
+            uc->transfer_file = share_filename;
+            free(uc->local_file);
+            uc->local_file = local_filename;
             uc->transferring = true;
             uc->transfer_pos = uc->transfer_start;
             uc->transfer_time = time(NULL);
             if (uc->transfer_time == (time_t) -1)
                 warn(_("Cannot get current time - %s\n"), errstr);
             break;
+        }
         case DC_MSG_CHECK_DOWNLOAD:
             msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_END);
             for (; uc->queue_pos < uc->info->download_queue->cur; uc->queue_pos++) {
@@ -639,16 +669,17 @@ user_result_fd_readable(DCUserConn *uc)
                     local_file = resolve_download_file(uc->info, queued);
                     uc->queued_valid = true;
                     uc->transfer_file = xstrdup(queued->filename);
+                    uc->local_file = xstrdup(local_file);
                     uc->occupied_slot = true;
                     queued->status = DC_QS_PROCESSING;
                     used_dl_slots++;
-                    msgq_put(uc->put_mq, MSGQ_STR, local_file, MSGQ_STR, uc->transfer_file, MSGQ_INT64, queued->length, MSGQ_END);
+                    msgq_put(uc->put_mq, MSGQ_STR, local_file, MSGQ_STR, uc->transfer_file, MSGQ_INT64, queued->length, MSGQ_INT, queued->flag, MSGQ_END);
                     FD_SET(uc->put_mq->fd, &write_fds);
                     free(local_file);
                     return;
                 }
             }
-            msgq_put(uc->put_mq, MSGQ_STR, NULL, MSGQ_STR, NULL, MSGQ_INT64, (uint64_t) 0, MSGQ_END);
+            msgq_put(uc->put_mq, MSGQ_STR, NULL, MSGQ_STR, NULL, MSGQ_INT64, (uint64_t) 0, MSGQ_INT, DC_TF_NORMAL, MSGQ_END);
             FD_SET(uc->put_mq->fd, &write_fds);
             break;
         case DC_MSG_DOWNLOAD_ENDED: {
@@ -663,22 +694,44 @@ user_result_fd_readable(DCUserConn *uc)
         case DC_MSG_CHECK_UPLOAD: {
             char *remote_file;
             char *local_file;
+            int type;
+            uint64_t size = 0;
+            DCTransferFlag flag = DC_TF_NORMAL;
+            bool permit_transfer = false;
 
-            msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_STR, &remote_file, MSGQ_END);
-            if (used_ul_slots >= my_ul_slots && !uc->info->slot_granted) {
-                uc->transfer_file = NULL;
-                uc->occupied_slot = false;
-                local_file = NULL;
-                free(remote_file);
+            msgq_get(uc->get_mq, MSGQ_INT, &id, MSGQ_INT, &type, MSGQ_STR, &remote_file, MSGQ_END);
+            local_file = resolve_upload_file(uc->info, type, remote_file, &flag, &size);
+            free(remote_file);
+            uc->transfer_file = NULL;
+            if (local_file != NULL) {
+                if (flag == DC_TF_LIST || (flag == DC_TF_NORMAL && size <= minislot_size)) {
+                    if (used_mini_slots < minislot_count) {
+                        used_mini_slots ++;
+                        uc->occupied_minislot = true;
+                        permit_transfer = true;
+                    } else if (used_ul_slots < my_ul_slots || uc->info->slot_granted) {
+                        used_ul_slots ++;
+                        uc->occupied_slot = true;
+                        permit_transfer = true;
+                    }
+                } else if (flag == DC_TF_NORMAL && size > minislot_size) {
+                    if (used_ul_slots < my_ul_slots || uc->info->slot_granted) {
+                        used_ul_slots ++;
+                        uc->occupied_slot = true;
+                        permit_transfer = true;
+                    }
+                }
+                if (permit_transfer) {
+                    uc->transfer_file = local_file;
+                } else {
+                    free(local_file);
+                    local_file = NULL;
+                }
             } else {
-                uc->transfer_file = remote_file;
-                uc->occupied_slot = true;
-                used_ul_slots++;
-                local_file = resolve_upload_file(uc->info, remote_file);
+                permit_transfer = true;
             }
-            msgq_put(uc->put_mq, MSGQ_BOOL, uc->occupied_slot, MSGQ_STR, local_file, MSGQ_END);
+            msgq_put(uc->put_mq, MSGQ_BOOL, permit_transfer, MSGQ_STR, local_file, MSGQ_END);
             FD_SET(uc->put_mq->fd, &write_fds);
-            free(local_file);
             break;
         }
         case DC_MSG_UPLOAD_ENDED: {
@@ -788,6 +841,9 @@ read_signal_input(void)
             } else if (child == parse_child) {
                 name = _("Parse process");
                 running = false;
+            } else if (child == update_child) {
+                name = _("FileList Update process");
+                running = false;
             } else {
                 /* Assume it was a child process */
                 name = _("User process");
@@ -843,23 +899,8 @@ run_script(const char *filename, bool allow_missing)
 }
 
 bool
-set_share_dir(const char *dir)
+add_share_dir(const char *dir)
 {
-    char *tmp = share_dir;
-    char sizebuf[LONGEST_HUMAN_READABLE+1];
-    /* set first, because we may be called as set_share_dir(share_dir); */
-    share_dir = (dir == NULL ? NULL : xstrdup(dir));
-    free(tmp);
-    if (!filelist_create(share_dir))
-    	return false;
-    if (hub_state >= DC_HUB_LOGGED_IN && !send_my_info())
-    	return false;
-    if (dir != NULL) {
-	screen_putf(_("Sharing %" PRIu64 " %s (%s) totally\n"),
-	    my_share_size,
-	    ngettext("byte", "bytes", my_share_size),
-	    human_readable(my_share_size, sizebuf, human_suppress_point_zero|human_autoscale|human_base_1024|human_SI|human_B, 1, 1));
-    }
     return true;
 }
 
@@ -867,17 +908,21 @@ void
 add_search_result(struct sockaddr_in *addr, char *results, uint32_t resultlen)
 {
     DCUDPMessage *msg;
+    /*
     int res = 0;
+    */
 
+    /*
     if (search_udpmsg_out->cur == 0) {
     	res = sendto(search_socket, results, resultlen, 0, (struct sockaddr *) addr, sizeof(struct sockaddr_in));
         if (res == 0 || (res < 0 && errno != EAGAIN)) {
-	    warn_socket_error(res, true, _("user (search result)"));
-	    return;
-	}
-	if (res == resultlen)
-	    return;
+	        warn_socket_error(res, true, _("user (search result)"));
+	        return;
+	    }
+	    if (res == resultlen)
+	        return;
     }
+    */
 
     msg = xmalloc(sizeof(DCUDPMessage)+resultlen);
     msg->addr = *addr;
@@ -904,7 +949,7 @@ search_input_available(void)
     res = byteq_recvfrom(search_recvq, search_socket, 0, (struct sockaddr *) &addr, &addrlen);
     if (res == 0 || (res < 0 && errno != EAGAIN)) {
     	warn_socket_error(res, false, _("user (search result)"));
-	return;
+        return;
     }
 
     dump_command(_("<=="), search_recvq->buf, search_recvq->cur);
@@ -917,22 +962,22 @@ static void
 search_now_writable(void)
 {
     while (search_udpmsg_out->cur > 0) {
-    	DCUDPMessage *msg = search_udpmsg_out->buf[0];
-	int res;
+        DCUDPMessage *msg = search_udpmsg_out->buf[0];
+        int res;
 
     	/* <Erwin> I don't think you can send off half a UDP packet, no.
     	 * Therefore we don't attempt to use non-blocking I/O on search_socket.
-	 */
-    	res = sendto(search_socket, msg->data, msg->len, 0, (struct sockaddr *) &msg->addr, sizeof(struct sockaddr_in));
-        if (res == 0 || (res < 0 && errno != EAGAIN))
-	    warn_socket_error(res, true, _("user (search result)"));
-	if (res < 0 && errno == EAGAIN)
-	    break;
-
+	    */
     	dump_command(_("==>"), msg->data, msg->len);
 
-	ptrv_remove_first(search_udpmsg_out);
-	free(msg);
+        res = sendto(search_socket, msg->data, msg->len, 0, (struct sockaddr *) &msg->addr, sizeof(struct sockaddr_in));
+        if (res == 0 || (res < 0 && errno != EAGAIN))
+	        warn_socket_error(res, true, _("user (search result)"));
+        if (res < 0 && errno == EAGAIN)
+            break;
+
+        ptrv_remove_first(search_udpmsg_out);
+        free(msg);
     }
 
     FD_CLR(search_socket, &write_fds);
@@ -997,48 +1042,52 @@ enable_active(uint16_t port)
     /* Create sockets. */
     listen_socket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
     if (listen_socket < 0) {
-	warn(_("Cannot create socket - %s\n"), errstr);
-	disable_active();
-	return false;
+        warn(_("Cannot create socket - %s\n"), errstr);
+        disable_active();
+        return false;
     }
 
     /* Set non-blocking */
     if (!fd_set_nonblock_flag(listen_socket, true)) {
     	warn(_("Cannot set non-blocking flag - %s\n"), errstr);
-	disable_active();
-	return false;
+        disable_active();
+        return false;
     }
 
     /* Enable listen address reusing */
     val = 1;
     if (setsockopt(listen_socket, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val)) < 0)
-	warn(_("Cannot enable address reusing - %s\n"), errstr);
+        warn(_("Cannot enable address reusing - %s\n"), errstr);
 
     /* Bind to address. If not binding, port is selected randomly. */
     addr_len = sizeof(addr);
     if (port != 0) {
     	addr.sin_family = AF_INET;
-	addr.sin_port = htons(port);
-	addr.sin_addr.s_addr = htonl(INADDR_ANY);
+        addr.sin_port = htons(port);
+        if (force_listen_addr.s_addr != INADDR_NONE) {
+            addr.sin_addr.s_addr = force_listen_addr.s_addr;
+        } else {
+            addr.sin_addr.s_addr = htonl(INADDR_ANY);
+        }
     	if (bind(listen_socket, (struct sockaddr *) &addr, addr_len) < 0) {
-	    warn(_("Cannot bind to address - %s\n"), errstr);
-	    disable_active();
-	    return false;
-	}
+            warn(_("Cannot bind to address - %s\n"), errstr);
+            disable_active();
+            return false;
+        }
     }
 
     /* Start listening for connections. */
     if (listen(listen_socket, LISTEN_QUEUE_CONNS) < 0) {
-	warn(_("Cannot listen - %s\n"), errstr);
-	disable_active();
-	return false;
+        warn(_("Cannot listen - %s\n"), errstr);
+        disable_active();
+        return false;
     }
 
     /* Find port we are listening on. */
     if (getsockname(listen_socket, (struct sockaddr *) &addr, &addr_len) < 0) {
     	warn(_("Cannot get socket address - %s\n"), errstr);
-	disable_active();
-	return false;
+        disable_active();
+        return false;
     }
 
     screen_putf(_("Listening on %s.\n"), sockaddr_in_str(&addr));
@@ -1051,18 +1100,18 @@ bool
 set_active(bool newactive, uint16_t port)
 {
     if (newactive) {
-	disable_active();
+        disable_active();
     	if (!enable_active(port))
-	    return false;
+	        return false;
     } else {
     	disable_active();
         listen_port = port;
     }
     /* Start of disable_search. */
     if (search_socket >= 0) {
-	if (close(search_socket) < 0)
-    	    warn(_("Cannot close socket - %s\n"), errstr);
-	search_socket = -1;
+	    if (close(search_socket) < 0)
+            warn(_("Cannot close socket - %s\n"), errstr);
+	    search_socket = -1;
     }
     /* End of disable_search. */
     enable_search();
@@ -1112,46 +1161,46 @@ main (int argc, char **argv)
     get_package_file("config", &config_file);
 
     while (true) {
-	c = getopt_long(argc, argv, short_opts, long_opts, NULL);
-	if (c == -1)
-	    break;
+	    c = getopt_long(argc, argv, short_opts, long_opts, NULL);
+	    if (c == -1)
+	        break;
 
-	switch (c) {
-	case 'c': /* --config */
-	    custom_config = true;
-	    free(config_file);
-	    config_file = xstrdup(optarg);
-	    break;
-	case 'n': /* --no-config */
-	    free(config_file);
-	    config_file = NULL;
-	    break;
-	case HELP_OPT: /* --help */
-	    printf(_("Usage: %s [OPTION]...\n"), quotearg(argv[0]));
-	    puts(_("Start microdc, a command-line based Direct Connect client.\n"));
-	    printf(_("  -n, --no-config  do not read config file on startup\n"));
-	    printf(_("      --help       display this help and exit\n"));
-	    printf(_("      --version    output version information and exit\n"));
-	    printf(_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
-	    exit(EXIT_SUCCESS);
-	case VERSION_OPT: /* --version */
-	    version_etc(stdout, NULL, PACKAGE, VERSION, "Oskar Liljeblad", NULL);
-	    exit(EXIT_SUCCESS);
+	    switch (c) {
+	    case 'c': /* --config */
+	        custom_config = true;
+	        free(config_file);
+	        config_file = xstrdup(optarg);
+	        break;
+	    case 'n': /* --no-config */
+	        free(config_file);
+	        config_file = NULL;
+	        break;
+	    case HELP_OPT: /* --help */
+	        printf(_("Usage: %s [OPTION]...\n"), quotearg(argv[0]));
+	        puts(_("Start microdc, a command-line based Direct Connect client.\n"));
+	        printf(_("  -n, --no-config  do not read config file on startup\n"));
+	        printf(_("      --help       display this help and exit\n"));
+	        printf(_("      --version    output version information and exit\n"));
+	        printf(_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
+	        exit(EXIT_SUCCESS);
+	    case VERSION_OPT: /* --version */
+	        version_etc(stdout, NULL, PACKAGE, VERSION, /*"Oskar Liljeblad",*/ "Vladimir Chugunov", NULL);
+	        exit(EXIT_SUCCESS);
         default:
             exit(EXIT_FAILURE);
-	}
+	    }
     }
     
     if (pipe(signal_pipe) < 0) {
     	warn(_("Cannot create pipe pair - %s\n"), errstr);
-	goto cleanup;
+	    goto cleanup;
     }
 
     main_process_id = getpid();
     sigact.sa_handler = signal_received;
     if (sigemptyset(&sigact.sa_mask) < 0) {
         warn(_("Cannot empty signal set - %s\n"), errstr);
-	goto cleanup;
+	    goto cleanup;
     }
     sigact.sa_flags = SA_RESTART;
 #ifdef HAVE_STRUCT_SIGACTION_SA_RESTORER
@@ -1160,26 +1209,17 @@ main (int argc, char **argv)
     /* Note: every signal registered with a non-ignore action here must
      * also be registered in user.c, either with an action or as ignored.
      */
-    if (sigaction(SIGINT, &sigact, NULL) < 0) {
+    if (sigaction(SIGINT,  &sigact, NULL) < 0 ||
+        sigaction(SIGTERM, &sigact, NULL) < 0 ||
+        sigaction(SIGUSR1, &sigact, NULL) < 0 ||
+        sigaction(SIGCHLD, &sigact, NULL) < 0) {
         warn(_("Cannot register signal handler - %s\n"), errstr);
-	goto cleanup;
-    }
-    if (sigaction(SIGTERM, &sigact, NULL) < 0) {
-        warn(_("Cannot register signal handler - %s\n"), errstr);
-	goto cleanup;
-    }
-    if (sigaction(SIGUSR1, &sigact, NULL) < 0) {
-        warn(_("Cannot register signal handler - %s\n"), errstr);
-	goto cleanup;
-    }
-    if (sigaction(SIGCHLD, &sigact, NULL) < 0) {
-        warn(_("Cannot register signal handler - %s\n"), errstr);
-	goto cleanup;
+	    goto cleanup;
     }
     sigact.sa_handler = SIG_IGN;
     if (sigaction(SIGPIPE, &sigact, NULL) < 0) {
         warn(_("Cannot register signal handler - %s\n"), errstr);
-	goto cleanup;
+	    goto cleanup;
     }
 
     FD_ZERO(&read_fds);
@@ -1193,6 +1233,10 @@ main (int argc, char **argv)
     hub_users = hmap_new();
     pending_userinfo = hmap_new();
 
+    set_main_charset("");
+    set_hub_charset("");
+    set_fs_charset("");
+
     user_conn_unknown_free = ptrv_new();
     delete_files = ptrv_new();
     delete_dirs = ptrv_new();
@@ -1207,8 +1251,8 @@ main (int argc, char **argv)
     download_dir = xstrdup(".");
     tmpdir = tempdir();
     if (tmpdir == NULL) {
-	warn(_("Cannot find directory for temporary files - %s\n"), errstr);
-	goto cleanup;
+	    warn(_("Cannot find directory for temporary files - %s\n"), errstr);
+	    goto cleanup;
     }
     {
         char *filename = xasprintf("%s.%d", PACKAGE, getpid());
@@ -1218,12 +1262,12 @@ main (int argc, char **argv)
     ptrv_append(delete_dirs, xstrdup(listing_dir));
     is_active = false;
     listen_port = 0;
+    if (!local_file_list_update_init())
+        goto cleanup;
     if (!set_active(false, listen_port))
     	goto cleanup;
     if (!enable_search())
     	goto cleanup;
-    if (!set_share_dir(NULL))
-    	goto cleanup;
     my_ul_slots = 3;
 
     if (!lookup_init())
@@ -1232,40 +1276,52 @@ main (int argc, char **argv)
         goto cleanup;
     command_init();
 
+    if (!local_file_list_init()) {
+        goto cleanup;
+    }
+
     if (config_file != NULL) {
         run_script(config_file, !custom_config);
-        config_file = NULL;
         free(config_file);
-    }      
+        config_file = NULL;
+    }
+
+    screen_prepare();
 
     while (running) {
     	fd_set res_read_fds;
-	fd_set res_write_fds;
+	    fd_set res_write_fds;
     	int res;
+        struct timeval tv;
+        tv.tv_sec = 1;
+        tv.tv_usec = 0;
 
-        screen_prepare();
-	res_read_fds = read_fds;
-	res_write_fds = write_fds;
-    	res = TEMP_FAILURE_RETRY(select(FD_SETSIZE, &res_read_fds, &res_write_fds, NULL, NULL));
-	if (res < 0) {
-	    warn(_("Cannot select - %s\n"), errstr);
-	    break;
-	}
+        screen_redisplay_prompt();
+
+	    res_read_fds = read_fds;
+	    res_write_fds = write_fds;
+    	res = TEMP_FAILURE_RETRY(select(FD_SETSIZE, &res_read_fds, &res_write_fds, NULL, &tv));
+	    if (res < 0) {
+	        warn(_("Cannot select - %s\n"), errstr);
+	        break;
+	    }
 
     	if (running && FD_ISSET(signal_pipe[0], &res_read_fds))
-	    read_signal_input();
+	        read_signal_input();
     	if (running && FD_ISSET(STDIN_FILENO, &res_read_fds))
     	    screen_read_input();
     	if (running && listen_socket >= 0 && FD_ISSET(listen_socket, &res_read_fds))
-	    handle_listen_connection();
-	if (running && hub_socket >= 0 && FD_ISSET(hub_socket, &res_read_fds))
-	    hub_input_available();
-	if (running && hub_socket >= 0 && FD_ISSET(hub_socket, &res_write_fds))
-	    hub_now_writable();
-	if (running && search_socket >= 0 && FD_ISSET(search_socket, &res_read_fds))
+	        handle_listen_connection();
+	    if (running && hub_socket >= 0 && FD_ISSET(hub_socket, &res_read_fds))
+	        hub_input_available();
+	    if (running && hub_socket >= 0 && FD_ISSET(hub_socket, &res_write_fds))
+	        hub_now_writable();
+        if (running)
+            check_hub_activity();
+	    if (running && search_socket >= 0 && FD_ISSET(search_socket, &res_read_fds))
     	    search_input_available();
-	if (running && search_socket >= 0 && FD_ISSET(search_socket, &res_write_fds))
-	    search_now_writable();
+	    if (running && search_socket >= 0 && FD_ISSET(search_socket, &res_write_fds))
+	        search_now_writable();
         if (running && FD_ISSET(lookup_request_mq->fd, &res_write_fds))
             lookup_request_fd_writable();
         if (running && FD_ISSET(lookup_result_mq->fd, &res_read_fds))
@@ -1274,18 +1330,23 @@ main (int argc, char **argv)
             parse_request_fd_writable();
         if (running && FD_ISSET(parse_result_mq->fd, &res_read_fds))
             parse_result_fd_readable();
-	if (running) {
-	    HMapIterator it;
+        if (running && FD_ISSET(update_request_mq->fd, &res_write_fds))
+            update_request_fd_writable();
+        if (running && FD_ISSET(update_result_mq->fd, &res_read_fds))
+            update_result_fd_readable();
 
-	    hmap_iterator(user_conns, &it);
-	    while (running && it.has_next(&it)) {
-	    	DCUserConn *uc = it.next(&it);
-	    	if (uc->put_mq != NULL && FD_ISSET(uc->put_mq->fd, &res_write_fds))
-	    	    user_request_fd_writable(uc);
-                if (uc->get_mq != NULL && FD_ISSET(uc->get_mq->fd, &res_read_fds))
-                    user_result_fd_readable(uc);
+	    if (running) {
+	        HMapIterator it;
+
+	        hmap_iterator(user_conns, &it);
+	        while (running && it.has_next(&it)) {
+	    	    DCUserConn *uc = it.next(&it);
+	    	    if (uc->put_mq != NULL && FD_ISSET(uc->put_mq->fd, &res_write_fds))
+	    	        user_request_fd_writable(uc);
+                    if (uc->get_mq != NULL && FD_ISSET(uc->get_mq->fd, &res_read_fds))
+                        user_result_fd_readable(uc);
+	        }
 	    }
-	}
     }
 
 cleanup:
@@ -1293,6 +1354,7 @@ cleanup:
     hub_disconnect();
     screen_finish();
     command_finish();
+    local_file_list_update_finish();
     file_list_parse_finish();
     lookup_finish();
 
@@ -1319,7 +1381,10 @@ cleanup:
     if (our_filelist != NULL)
     	filelist_free(our_filelist);
 
-    free(hub_charset);
+    set_main_charset(NULL);
+    set_hub_charset(NULL);
+    set_fs_charset(NULL);
+
     free(hub_name);
     free(my_nick);
     free(my_description);
@@ -1329,45 +1394,40 @@ cleanup:
     free(download_dir);
     free(listing_dir);
 
-    if (to_hub_iconv != (iconv_t) -1 && iconv_close(to_hub_iconv) != 0)
-        warn(_("cannot close character convertor - %s\n"), errstr);
-    if (from_hub_iconv != (iconv_t) -1 && iconv_close(from_hub_iconv) != 0)
-        warn(_("cannot close character convertor - %s\n"), errstr);
-
     if (delete_files != NULL) {
-	for (c = 0; c < delete_files->cur; c++) {
-    	    char *filename = delete_files->buf[c];
-    	    struct stat st;
+	    for (c = 0; c < delete_files->cur; c++) {
+    	        char *filename = delete_files->buf[c];
+    	        struct stat st;
 
-	    if (stat(filename, &st) < 0) {
-		if (errno != ENOENT)
-	    	    warn(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
-		free(filename);
-		continue;
-	    }
+	        if (stat(filename, &st) < 0) {
+		        if (errno != ENOENT)
+	    	        warn(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
+		        free(filename);
+		        continue;
+	        }
     	    if (unlink(filename) < 0)
-		warn(_("%s: Cannot remove file - %s\n"), quotearg(filename), errstr);
-	    free(filename);
-	}
-	ptrv_free(delete_files);
+		        warn(_("%s: Cannot remove file - %s\n"), quotearg(filename), errstr);
+	        free(filename);
+	    }
+	    ptrv_free(delete_files);
     }
 
     if (delete_dirs != NULL) {
-	for (c = 0; c < delete_dirs->cur; c++) {
-    	    char *filename = delete_dirs->buf[c];
-    	    struct stat st;
+	    for (c = 0; c < delete_dirs->cur; c++) {
+    	        char *filename = delete_dirs->buf[c];
+    	        struct stat st;
 
-	    if (stat(filename, &st) < 0) {
-		if (errno != ENOENT)
-	    	    warn(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
-		free(filename);
-		continue;
-	    }
+	        if (stat(filename, &st) < 0) {
+		        if (errno != ENOENT)
+	    	        warn(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
+		        free(filename);
+		        continue;
+	        }
     	    if (rmdir(filename) < 0)
-		warn(_("%s: Cannot remove file - %s\n"), quotearg(filename), errstr);
-	    free(filename);
-	}
-	ptrv_free(delete_dirs);
+		        warn(_("%s: Cannot remove file - %s\n"), quotearg(filename), errstr);
+	        free(filename);
+	    }
+	    ptrv_free(delete_dirs);
     }
 
     if (search_socket >= 0 && close(search_socket) < 0)
