@@ -85,6 +85,7 @@ typedef enum {
     DC_USER_MYNICK = 35,	/* Waiting for $MyNick */
     DC_USER_LOCK,  	    	/* Waiting for $Lock */
     DC_USER_DIRECTION,     	/* Waiting for $Direction */
+    DC_USER_SUPPORTS,     	/* Waiting for $Supports */
     DC_USER_KEY,                /* Waiting for $Key */
     DC_USER_GET,   	    	/* Waiting for $Get (only when sending) */
     DC_USER_SEND_GET,  	    	/* Waiting for $Send or $Get (only when sending) */
@@ -110,6 +111,11 @@ typedef enum {
     DC_TYPE_REG,
 } DCFileType;
 
+typedef enum {
+    DC_LS_LONG_MODE = 1,
+    DC_LS_TTH_MODE  = 2,
+} DCLsMode;
+
 typedef enum {
     DC_MSG_SCREEN_PUT,
     DC_MSG_WANT_DOWNLOAD,
@@ -144,6 +150,12 @@ typedef enum {
     DC_QS_ERROR,	/* error occured during download */
 } DCQueuedStatus;
 
+typedef enum {
+    DC_ADCGET_FILE,	/* Upload by filename */
+    DC_ADCGET_TTH,	/* Upload by file root */
+    DC_ADCGET_TTHL	/* Upload tth leaves */
+} DCAdcgetType;
+
 typedef struct _DCUserConn DCUserConn;
 typedef struct _DCUserInfo DCUserInfo;
 typedef struct _DCFileList DCFileList;
@@ -268,11 +280,14 @@ struct _DCFileList {
     uint64_t size;	/* total size of all contained files for DC_TYPE_DIR */
     union {
     	struct {
-	    /*char tth[39];*/
-	} reg;
-	struct {
-	    HMap *children;
-	} dir;
+            char    has_tth;
+	        char    tth[39];
+            time_t  mtime;
+	    } reg;
+	    struct {
+	        char *real_path;
+	        HMap *children;
+	    } dir;
     };
 };
 
@@ -315,9 +330,11 @@ struct _DCUserConn {
                          * a message just received on get_mq.
                          */
     bool occupied_slot; /* true if used_*_slots were increased for this user connection */
+    bool occupied_minislot; /* true if used_mini_slots were increased for this user connection */
     uint32_t queue_pos;
     bool queued_valid;		/* true unless the QueuedFile by queue_pos has been removed */
     char *transfer_file;
+    char *local_file;
     bool transferring;
     uint64_t transfer_start;	/* the byte where the transfer started */
     uint64_t transfer_pos;	/* current byte position */
@@ -343,9 +360,6 @@ extern char *log_filename;
 extern DCHubState hub_state;
 extern ByteQ *hub_recvq;
 extern ByteQ *hub_sendq;
-extern char *hub_charset;
-extern iconv_t to_hub_iconv;
-extern iconv_t from_hub_iconv;
 extern HMap *hub_users;
 extern HMap *user_conns;
 extern struct sockaddr_in local_addr;
@@ -362,17 +376,20 @@ extern char *my_nick;
 extern char *my_description;
 extern char *my_speed;
 extern char *my_email;
-extern char *share_dir;
 extern uint64_t my_share_size;
 extern char *download_dir;
 extern char *listing_dir;
 extern bool is_active;
+extern bool auto_reconnect;
 extern uint32_t my_ul_slots;
 extern fd_set read_fds;
 extern fd_set write_fds;
 extern char *my_password;
 extern PtrV *delete_files;  /* XXX: use LList? */
 extern PtrV *delete_dirs;   /* XXX: use LList? */
+extern uint32_t minislot_count;
+extern uint32_t minislot_size;
+extern int used_mini_slots;
 extern int used_ul_slots;
 extern int used_dl_slots;
 
@@ -394,7 +411,9 @@ void user_info_free(DCUserInfo *ui);
 void hub_new(const char *hostname, uint16_t port);
 void hub_connect(struct sockaddr_in *addr);
 void hub_disconnect(void);
+void hub_reconnect(void);
 void hub_now_writable(void);
+void check_hub_activity();
 bool send_my_info(void);
 extern struct sockaddr_in hub_addr;
 extern char *hub_name;
@@ -402,6 +421,8 @@ void say_user_completion_generator(DCCompletionInfo *ci);
 void user_completion_generator(DCCompletionInfo *ci);
 void user_or_myself_completion_generator(DCCompletionInfo *ci);
 void user_with_queue_completion_generator(DCCompletionInfo *ci);
+void hub_reload_users();
+void hub_set_connected(bool state);
 
 /* command.c */
 void command_execute(const char *line);
@@ -419,6 +440,7 @@ void screen_finish(void);
 void screen_suspend(void);
 void screen_wakeup(bool print_newline_first);
 void screen_prepare(void);
+void screen_redisplay_prompt();
 void screen_read_input(void);
 void screen_get_size(int *rows, int *cols);
 void set_screen_prompt(const char *prompt, ...) __attribute__ ((format (printf, 1, 2)));
@@ -445,7 +467,8 @@ void user_main(int get_fd[2], int put_fd[2], struct sockaddr_in *addr, int sock)
 /* main.c */
 /*bool get_user_conn_status(DCUserConn *uc);*/
 bool set_active(bool newactive, uint16_t newport);
-bool set_share_dir(const char *dir);
+bool add_share_dir(const char *dir);
+bool del_share_dir(const char *dir);
 bool has_user_conn(DCUserInfo *info, DCTransferDirection dir);
 uint32_t get_user_conn_count(DCUserInfo *info);
 bool get_package_file(const char *name, char **outname);
@@ -463,12 +486,14 @@ extern uint64_t bytes_sent;
 
 /* fs.c */
 DCFileList *new_file_node(const char *name, DCFileType type, DCFileList *parent);
+void rename_node(DCFileList *node, const char* new_name);
+void set_child_node(DCFileList *parent, DCFileList *child);
 void filelist_free(DCFileList *fl);
-void filelist_list(DCFileList *fl, bool longmode);
+void filelist_list(DCFileList *fl, int mode);
 DCFileList *filelist_lookup(DCFileList *node, const char *filename);
 char *filelist_get_path(DCFileList *node);
 char *filelist_get_path_with_trailing_slash(DCFileList *node);
-char *resolve_upload_file(DCUserInfo *ui, const char *name);
+char *resolve_upload_file(DCUserInfo *ui, DCAdcgetType ul_type, const char *name, DCTransferFlag* flag, uint64_t* size);
 char *resolve_download_file(DCUserInfo *ui, DCQueuedFile *queued);
 bool filelist_create(const char *dir);
 char *translate_local_to_remote(const char *localname);
@@ -480,10 +505,19 @@ void remote_path_completion_generator(DCCompletionInfo *ci);
 void remote_dir_completion_generator(DCCompletionInfo *ci);
 char *apply_cwd(const char *path); /* XXX: move transfer.c? */
 extern DCFileList *our_filelist;
+extern time_t      our_filelist_last_update;
 void filelist_list_recursively(DCFileList *node, char *basepath);
 void remote_wildcard_expand(char *matchpath, bool *quotedptr, const char *basedir, DCFileList *basenode, PtrV *results);
 bool has_leading_slash(const char *str);
+void dir_to_filelist(DCFileList *parent, const char *path);
+bool write_filelist_file(DCFileList* root, const char* prefix);
     
+/* xml_flist.c */
+int write_xml_filelist(int fd, DCFileList* root);
+int write_bzxml_filelist(int fd, DCFileList* root);
+DCFileList* filelist_xml_open(const char* filename);
+DCFileList* filelist_bzxml_open(const char* filename);
+
 /* connection.c */
 char *decode_lock(const char *lock, size_t locklen, uint32_t basekey);
 char *escape_message(const char *str);
@@ -493,15 +527,15 @@ void dump_command(const char *header, const char *buf, size_t len);
 /* util.c */
 int safe_rename(const char *oldpath, const char *newpath);
 int ilog10(uint64_t c);
-int mkdirs_for_file(char *filename, bool deletedirs);
+int mkdirs_for_file(char *filename);
 char *catfiles(const char *p1, const char *p2);
 char *catfiles_with_trailing_slash(const char *p1, const char *p2);
 bool fd_set_status_flags(int fd, bool set, int modflags);
 #define fd_set_nonblock_flag(f,s) fd_set_status_flags(f,s,O_NONBLOCK)
 char *getenv_default(const char *name, char *defvalue);
-#define IS_CURRENT_DIR(x) ((x)[0] == '.' && (x)[1] == '\0')
-#define IS_PARENT_DIR(x)  ((x)[0] == '.' && (x)[1] == '.' && (x)[2] == '\0')
-#define IS_SPECIAL_DIR(x) (IS_CURRENT_DIR(x) || IS_PARENT_DIR(x))
+
+#include "tth_file.h"
+
 #define COMPARE_RETURN(a,b) { if ((a) < (b)) return -1; if ((a) > (b)) return 1; }
 #define COMPARE_RETURN_FUNC(f) { int _c = (f); if (_c != 0) return _c; }
 char *sockaddr_in_str(struct sockaddr_in *addr);
@@ -517,13 +551,16 @@ struct dirent *xreaddir(DIR *dh);
 char *elapsed_time_to_string(time_t elapsed, char *buf);
 
 /* search.c */
-bool parse_search_selection(char *str, DCSearchSelection *data);
+int parse_search_selection(char *str, DCSearchSelection *data);
 bool perform_inbound_search(DCSearchSelection *data, DCUserInfo *ui, struct sockaddr_in *addr);
 extern PtrV *our_searches;
 bool add_search_request(char *args);
 void handle_search_result(char *buf, uint32_t len);
 void free_search_request(DCSearchRequest *sr);
 char *search_selection_to_string(DCSearchSelection *sr);
+void search_string_new(DCSearchString *sp, const char *p, int len);
+void search_hash_new(DCSearchString *sp, const char *p, int len);
+void search_string_free(DCSearchString *sp);
 
 /* variables.c */
 void cmd_set(int argc, char **argv);
@@ -550,5 +587,61 @@ DCFileListParse *add_parse_request(DCFileListParseCallback callback, const char
 void cancel_parse_request(DCFileListParse *parse);
 void parse_result_fd_readable(void);
 void parse_request_fd_writable(void);
+void* data_to_filelist(void *dataptr, DCFileList **outnode);
+void  filelist_to_data(DCFileList *node, void **dataptr, size_t *sizeptr);
+
+/* local_flist.c */
+extern MsgQ *update_request_mq;
+extern MsgQ *update_result_mq;
+extern pid_t update_child;
+extern char* update_status;
+extern time_t filelist_refresh_timeout;
+bool local_file_list_update_init(void);
+bool local_file_list_init(void);
+void local_file_list_update_finish(void);
+bool update_request_add_shared_dir(const char* dir);
+bool update_request_del_shared_dir(const char* dir);
+bool update_request_set_listing_dir(const char* dir);
+bool update_request_set_hub_charset(const char* charset);
+bool update_request_set_fs_charset(const char* charset);
+bool update_request_set_filelist_refresh_timeout(time_t seconds);
+/*
+DCFileListParse *add_parse_request(DCFileListParseCallback callback, const char *filename, void *userdata);
+void cancel_parse_request(DCFileListParse *parse);
+*/
+void update_result_fd_readable(void);
+void update_request_fd_writable(void);
+
+/* charsets.c */
+#include "charsets.h"
+EXPORT_CHARSET(main);
+EXPORT_CHARSET(hub);
+EXPORT_CHARSET(fs);
+EXPORT_CHARSET(log);
+EXPORT_CONST_CHARSET(utf8);
+
+void set_main_charset(const char* charset);
+void set_hub_charset(const char* charset);
+void set_fs_charset(const char* charset);
+void set_log_charset(const char* charset);
+
+EXPORT_ICONV_CONVERSION(main, hub);
+EXPORT_ICONV_CONVERSION(hub, main);
+EXPORT_ICONV_CONVERSION(main, fs);
+EXPORT_ICONV_CONVERSION(fs, main);
+EXPORT_ICONV_CONVERSION(hub, fs);
+EXPORT_ICONV_CONVERSION(fs, hub);
+EXPORT_ICONV_CONVERSION(main, log);
+EXPORT_ICONV_CONVERSION(log, main);
+
+#if defined(HAVE_LIBXML2)
+EXPORT_ICONV_CONVERSION(utf8, fs);
+EXPORT_ICONV_CONVERSION(fs, utf8);
+EXPORT_ICONV_CONVERSION(utf8, main);
+EXPORT_ICONV_CONVERSION(main, utf8);
+EXPORT_ICONV_CONVERSION(utf8, hub);
+EXPORT_ICONV_CONVERSION(hub, utf8);
+#endif
+
 
 #endif
