@@ -38,6 +38,7 @@
 #include "xvasprintf.h"		/* Gnulib */
 #include "quotearg.h"		/* Gnulib */
 #include "minmax.h"		/* Gnulib */
+#include "iconvme.h"
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
@@ -76,7 +77,7 @@ static void flag_vputf(DCDisplayFlag flag, const char *format, va_list args);
 
 char *log_filename = NULL;
 ScreenWriter screen_writer = flag_vputf;
-static char *screen_prompt;
+static char *screen_prompt = NULL;
 static FILE *log_fh = NULL;
 static PtrV *suspend_msgs = NULL;
 static ScreenState screen_state = SCREEN_UNINITIALIZED;
@@ -227,6 +228,10 @@ screen_warn_writer(const char *format, va_list args)
 static void
 flag_vputf(DCDisplayFlag flag, const char *format, va_list args)
 {
+    //va_list args2;
+
+    //va_copy(args2, args);
+
     if (display_flags & flag) {
         if (screen_state == SCREEN_SUSPENDED) {
             ptrv_append(suspend_msgs, xvasprintf(format, args));
@@ -236,10 +241,24 @@ flag_vputf(DCDisplayFlag flag, const char *format, va_list args)
                 screen_state = SCREEN_RL_CLEARED;
             }
             vprintf(format, args);
+            fflush(stdout);
         }
     }
-    if (log_fh != NULL && log_flags & flag)
-        vfprintf(log_fh, format, args);
+    if (log_fh != NULL && log_flags & flag) {
+        char c_time[1024];
+        time_t now = time(NULL);
+        struct tm _tm = {0};
+        if (NULL != localtime_r(&now, &_tm) && 0 != strftime(c_time, 1023, "%d.%m.%Y %H:%M:%S", &_tm)) {
+            fprintf(log_fh, "%s ", c_time);
+        }
+        char* msg = xvasprintf(format, args);
+        //va_end(args2);
+        char* log_msg = main_to_log_string(msg);
+        free(msg);
+        fprintf(log_fh, log_msg);
+        free(log_msg);
+        fflush(log_fh);
+    }
 }
 
 void
@@ -261,8 +280,16 @@ user_input(char *line)
     /* Readline has already made way for us. */
     screen_state = SCREEN_RL_CLEARED;
 
-    if (log_fh != NULL)
-	fprintf(log_fh, "> %s\n", line == NULL ? "(null)" : line);
+    if (log_fh != NULL) {
+        char c_time[1024];
+        time_t now = time(NULL);
+        struct tm _tm = {0};
+        if (NULL != localtime_r(&now, &_tm) && 0 != strftime(c_time, 1023, "%d.%m.%Y %H:%M:%S", &_tm)) {
+            fprintf(log_fh, "%s ", c_time);
+        }
+	    fprintf(log_fh, "> %s\n", line == NULL ? "(null)" : line);
+        fflush(log_fh);
+    }
     
     if (line == NULL) {
         /* Ctrl+D was pressed on an empty line. */
@@ -356,7 +383,7 @@ screen_finish(void)
         rl_callback_handler_remove();
         if (screen_state == SCREEN_RL_DISPLAYED)
             putchar('\n');
-	FD_CLR(STDIN_FILENO, &read_fds);
+        FD_CLR(STDIN_FILENO, &read_fds);
     }
 
     if (screen_state >= SCREEN_NO_HANDLER) {
@@ -364,19 +391,19 @@ screen_finish(void)
 
 	/* Save history */
     	get_package_file("history", &path);
-    	if (mkdirs_for_file(path, false) >= 0) {
-	    if (write_history(path) != 0)
-		warn(_("%s: Cannot write history - %s\n"), quotearg(path), errstr);
-	}
-	free(path);
+    	if (mkdirs_for_file(path) >= 0) {
+            if (write_history(path) != 0)
+                warn(_("%s: Cannot write history - %s\n"), quotearg(path), errstr);
+        }
+        free(path);
 
         rl_pre_input_hook = NULL;
 
-	warn_writer = default_warn_writer;
+        warn_writer = default_warn_writer;
         screen_state = SCREEN_UNINITIALIZED;
         free(screen_prompt);
 
-	set_log_file(NULL, false);
+        set_log_file(NULL, false);
     }
 }
 
@@ -554,30 +581,41 @@ screen_prepare(void)
     if (screen_state == SCREEN_UNINITIALIZED) {
     	char *path;
 
-    	screen_state = SCREEN_NO_HANDLER;
-	warn_writer = screen_warn_writer;
-	screen_prompt = xasprintf("%s> ", PACKAGE);
+        screen_state = SCREEN_NO_HANDLER;
+        warn_writer = screen_warn_writer;
+        if (screen_prompt == NULL)
+            screen_prompt = xasprintf("%s> ", PACKAGE);
 
-	rl_readline_name = PACKAGE;
-	rl_add_defun(PACKAGE "-complete", completion_readline, '\t');
-    	rl_pre_input_hook = fix_winch;
+        rl_readline_name = PACKAGE;
+        rl_add_defun(PACKAGE "-complete", completion_readline, '\t');
+        rl_pre_input_hook = fix_winch;
 
         using_history();
-    	get_package_file("history", &path);
-	if (read_history(path) != 0 && errno != ENOENT)
-    	    warn(_("%s: Cannot read history - %s\n"), quotearg(path), errstr);
+        get_package_file("history", &path);
+        if (path != NULL) {
+            if (read_history(path) != 0 && errno != ENOENT) {
+                warn(_("%s: Cannot read history - %s\n"), quotearg(path), errstr);
+            }
+            free(path);
+        }
     }
     if (screen_state == SCREEN_NO_HANDLER) {
         rl_callback_handler_install(screen_prompt, user_input);
         FD_SET(STDIN_FILENO, &read_fds);
-    }
-    else if (screen_state == SCREEN_RL_CLEARED) {
+    } else if (screen_state == SCREEN_RL_CLEARED) {
         rl_set_prompt(screen_prompt);
         rl_redisplay();
     }
     screen_state = SCREEN_RL_DISPLAYED;
 }
 
+void screen_redisplay_prompt()
+{
+    rl_set_prompt(screen_prompt);
+    rl_redisplay();
+    screen_state = SCREEN_RL_DISPLAYED;
+}
+
 /* This function is called from the main loop when there's input for us to
  * read on stdin.
  */
