@@ -36,6 +36,14 @@
 #include "common/intutil.h"
 #include "microdc.h"
 
+//#define _TRACE
+#if defined(_TRACE)
+#include <stdio.h>
+#define TRACE(x) printf x; fflush(stdout)
+#else
+#define TRACE(x)
+#endif
+
 #define MAX_RESULTS_ACTIVE 10		/* Max number of search results to send to active users */
 #define MAX_RESULTS_PASSIVE 5		/* Max number of search results to send to passive users */
 
@@ -69,14 +77,32 @@ compare_search_response(DCSearchResponse *sr1, DCSearchResponse *sr2)
     return 0;
 }
 
-static void
+void
 search_string_new(DCSearchString *sp, const char *p, int len)
+{
+    int c;
+    unsigned char *u_str;
+
+    u_str = xstrndup(p, len);
+    for (c = 0; c < len; c++)
+    	u_str[c] = tolower(u_str[c]);
+    sp->len = len;
+    sp->str = u_str;
+
+    for (c = 0; c < 256; c++)
+    	sp->delta[c] = len+1;
+    for (c = 0; c < len; c++)
+    	sp->delta[ u_str[c] ] = len-c;
+}
+
+void
+search_hash_new(DCSearchString *sp, const char *p, int len)
 {
     int c;
 
     sp->str = xstrndup(p, len);
     for (c = 0; c < len; c++)
-    	sp->str[c] = tolower(sp->str[c]);
+    	sp->str[c] = toupper(sp->str[c]);
     sp->len = len;
 
     for (c = 0; c < 256; c++)
@@ -85,6 +111,31 @@ search_string_new(DCSearchString *sp, const char *p, int len)
     	sp->delta[(uint8_t) sp->str[c]] = len-c;
 }
 
+void
+search_string_free(DCSearchString *sp)
+{
+    free(sp->str);
+}
+
+static bool
+parse_hash(char *str, DCSearchSelection *ss)
+{
+    int len = strlen(str);
+    flag_putf(DC_DF_DEBUG, _("incoming hash: %s\n"), str);
+    if (len > 4 && str[3] == ':' &&
+        (str[0] == 't' || str[0] == 'T') && 
+        (str[1] == 't' || str[1] == 'T') && 
+        (str[2] == 'h' || str[2] == 'H')) {
+        /* TTH hash lookup */
+
+        ss->patterncount = 1;
+        ss->patterns = xmalloc(sizeof(DCSearchString));
+        search_hash_new(ss->patterns, str+4, len-4);
+        return true;
+    }
+    return false;
+}
+
 static bool
 parse_search_strings(char *str, DCSearchSelection *ss)
 {
@@ -99,8 +150,9 @@ parse_search_strings(char *str, DCSearchSelection *ss)
     }
     if (*t1)
         ss->patterncount++;
+
     if (ss->patterncount == 0)
-	return false;
+	    return false;
 
     ss->patterns = xmalloc(sizeof(DCSearchString) * ss->patterncount);
 
@@ -108,8 +160,8 @@ parse_search_strings(char *str, DCSearchSelection *ss)
     for (t1 = str; (t2 = strchr(t1, '$')) != NULL; t1 = t2+1) {
     	if (t2 != str && t2[-1] != '$') {
     	    search_string_new(ss->patterns+c, t1, t2-t1);
-	    c++;
-	}
+	        c++;
+	    }
     }
     if (*t1)
        search_string_new(ss->patterns+c, t1, strlen(t1));
@@ -143,7 +195,7 @@ match_file_extension(const char *filename, DCSearchDataType type)
     return false;
 }
 
-bool
+int
 parse_search_selection(char *str, DCSearchSelection *data)
 {
     char sizeres;
@@ -151,53 +203,65 @@ parse_search_selection(char *str, DCSearchSelection *data)
     char *sizestr;
     uint64_t size;
     char *datatype;
+    char *tmpstr;
+    bool ret;
 
     if (str[0] != 'T' && str[0] != 'F')
-    	return false;
+    	return 0;
     if (str[1] != '?')
-    	return false;
+    	return 0;
     if (str[2] != 'T' && str[2] != 'F')
-    	return false;
+    	return 0;
     if (str[3] != '?')
-    	return false;
+    	return 0;
     sizeres = str[0];
     sizemin = str[2];
 
     str += 4;
     sizestr = strsep(&str, "?");
     if (sizestr == NULL)
-    	return false;
+    	return 0;
     if (!parse_uint64(sizestr, &size))
-    	return false;
+    	return 0;
     datatype = strsep(&str, "?");
     if (datatype == NULL || datatype[0] < '1' || datatype[0] > '9' || datatype[1] != '\0')
-    	return false;
+    	return 0;
     if (*str == '\0')
-    	return false;
+    	return 2;
     if (strlen(str) >= 1 << 16) /* Needed for delta match */
-    	return false;
+    	return 0;
 
     if (sizeres) {
-    	if (sizemin) {
-    	    data->size_min = size;
-    	    data->size_max = UINT64_MAX;
-	} else {
-    	    data->size_min = 0;
-    	    data->size_max = size;
-	}
+        if (sizemin) {
+            data->size_min = size;
+            data->size_max = UINT64_MAX;
+        } else {
+            data->size_min = 0;
+            data->size_max = size;
+        }
     } else {
-    	data->size_min = 0;
-    	data->size_max = UINT64_MAX;
+        data->size_min = 0;
+        data->size_max = UINT64_MAX;
     }
     data->datatype = datatype[0]-'1';
 
-    return parse_search_strings(str, data);
+    if (data->datatype == DC_SEARCH_CHECKSUM) {
+        return parse_hash(str, data);
+    }
+
+    /* convert search string from hub charset to local charset */
+    tmpstr = hub_to_main_string(str);
+    ret = parse_search_strings(tmpstr, data);
+    free (tmpstr);
+
+    return ret;
 }
 
 static bool
 match_search_pattern(const char *t, DCSearchString *pattern)
 {
     const char *end;
+    const unsigned char *ut;
     uint32_t tlen;
 
     tlen = strlen(t);
@@ -207,10 +271,12 @@ match_search_pattern(const char *t, DCSearchString *pattern)
     end = t + tlen - pattern->len + 1;
     while (t < end) {
     	uint32_t i = 0;
-	for (; pattern->str[i] && pattern->str[i] == tolower(t[i]); i++);
+	ut = (const unsigned char *)t;
+	for (; pattern->str[i] && pattern->str[i] == (char)tolower(ut[i]); i++)
+	    ;
 	if (pattern->str[i] == '\0')
 	    return true;
-	t += pattern->delta[(uint8_t) tolower(t[pattern->len])];
+	t += pattern->delta[(uint8_t) tolower(ut[pattern->len])];
     }
     return false;
 }
@@ -232,6 +298,10 @@ append_result(DCFileList *node, DCUserInfo *ui, struct sockaddr_in *addr)
     StrBuf *sb;
     char *lpath;
     char *rpath;
+    char *conv_rpath;
+    char *hub_my_nick;
+    char *hub_hub_name;
+    char *hub_ui_nick;
     int free_slots;
 
     sb = strbuf_new();
@@ -241,14 +311,30 @@ append_result(DCFileList *node, DCUserInfo *ui, struct sockaddr_in *addr)
 
     free_slots = used_ul_slots > my_ul_slots ? 0 : my_ul_slots-used_ul_slots;
 
-    strbuf_appendf(sb, "$SR %s %s", my_nick, rpath);
+    conv_rpath = main_to_hub_string(rpath);
+    hub_my_nick = main_to_hub_string(my_nick);
+    hub_hub_name = main_to_hub_string(hub_name);
+    if ( ui != NULL)
+        hub_ui_nick = main_to_hub_string(ui->nick);
+    else
+        hub_ui_nick = NULL;
+
+    strbuf_appendf(sb, "$SR %s %s", hub_my_nick, conv_rpath);
+
     if (node->type == DC_TYPE_REG)
     	strbuf_appendf(sb, "\x05%" PRIu64, node->size);
-    strbuf_appendf(sb, " %d/%d\x05%s (%s)",
-    	    free_slots, my_ul_slots, hub_name,
-	    sockaddr_in_str(&hub_addr));
+    strbuf_appendf(sb, " %d/%d\x05", free_slots, my_ul_slots);
+    if (node->type == DC_TYPE_REG && node->reg.has_tth) {
+        unsigned char tth[40];
+        memcpy(tth, node->reg.tth, sizeof(node->reg.tth));
+        tth[39] = '\0';
+        strbuf_appendf(sb, "TTH:%s", tth);
+    } else {
+        strbuf_appendf(sb, "%s", hub_hub_name);
+    }
+    strbuf_appendf(sb, " (%s)", sockaddr_in_str(&hub_addr));
     if (ui != NULL)
-    	strbuf_appendf(sb, "\x05%s", ui->nick);
+    	strbuf_appendf(sb, "\x05%s", hub_ui_nick);
     strbuf_append(sb, "|");
 
     if (ui != NULL) {
@@ -257,6 +343,10 @@ append_result(DCFileList *node, DCUserInfo *ui, struct sockaddr_in *addr)
     	add_search_result(addr, sb->buf, strbuf_length(sb));
     }
     
+    free(hub_ui_nick);
+    free(hub_hub_name);
+    free(hub_my_nick);
+    free(conv_rpath);   
     strbuf_free(sb);
 }
 
@@ -266,18 +356,26 @@ filelist_search(DCFileList *node, DCSearchSelection *data, int maxresults, DCUse
     assert(maxresults > 0);
 
     if (node->type == DC_TYPE_REG) {
-    	if (data->datatype == DC_SEARCH_FOLDERS || data->datatype == DC_SEARCH_CHECKSUM) /* TTH not supported yet */
-	    return 0;
-    	if (node->size < data->size_min)
-	    return 0;
-	if (node->size > data->size_max)
-	    return 0;
-	if (!match_search_patterns(node->name, data))
-	    return 0;
-	if (!match_file_extension(node->name, data->datatype))
-	    return 0;
-	append_result(node, ui, addr);
-	return 1;
+        if (data->datatype == DC_SEARCH_CHECKSUM) {
+            if (!node->reg.has_tth)
+                return 0;
+            if (strncmp(node->reg.tth, data->patterns->str, sizeof(node->reg.tth)) != 0) { /* only TTH is supported up to now */
+                return 0;
+            }
+        } else {
+    	    if (data->datatype == DC_SEARCH_FOLDERS)
+	            return 0;
+    	    if (node->size < data->size_min)
+	            return 0;
+	        if (node->size > data->size_max)
+	            return 0;
+	        if (!match_search_patterns(node->name, data))
+	            return 0;
+	        if (!match_file_extension(node->name, data->datatype))
+	            return 0;
+        }
+	    append_result(node, ui, addr);
+	    return 1;
     }
 
     if (node->type == DC_TYPE_DIR) {
@@ -285,23 +383,24 @@ filelist_search(DCFileList *node, DCSearchSelection *data, int maxresults, DCUse
     	int curresults = 0;
 
     	if (data->datatype == DC_SEARCH_ANY || data->datatype == DC_SEARCH_FOLDERS) {
-	    if (match_search_patterns(node->name, data)) {
-    	    	append_result(node, ui, addr);
-    	    	curresults++;
-		if (curresults >= maxresults)
-	    	    return curresults;
+	        if (match_search_patterns(node->name, data)) {
+                append_result(node, ui, addr);
+            	curresults++;
+		        if (curresults >= maxresults)
+                    return curresults;
+	        }
 	    }
-	}
 
-	hmap_iterator(node->dir.children, &it);
-	while (it.has_next(&it)) {
-    	    DCFileList *subnode = it.next(&it);
-	    curresults += filelist_search(subnode, data, maxresults-curresults, ui, addr);
-	    if (curresults >= maxresults)
-    	    	break;
+	    hmap_iterator(node->dir.children, &it);
+	    while (it.has_next(&it)) {
+            DCFileList *subnode = it.next(&it);
+	        curresults += filelist_search(subnode, data, maxresults-curresults, ui, addr);
+            if ((data->datatype == DC_SEARCH_CHECKSUM && curresults > 0) ||
+	            (curresults >= maxresults))
+                break;
     	}
 	
-	return curresults;
+	    return curresults;
     }
 
     return 0;
@@ -314,6 +413,8 @@ perform_inbound_search(DCSearchSelection *data, DCUserInfo *ui, struct sockaddr_
     int maxresults;
     int curresults;
 
+    TRACE(("%s:%d: \n", __FUNCTION__, __LINE__));
+
     maxresults = (ui == NULL ? MAX_RESULTS_ACTIVE : MAX_RESULTS_PASSIVE);
     curresults = filelist_search(our_filelist, data, maxresults, ui, addr);
 
@@ -343,6 +444,7 @@ parse_search_response(char *buf, uint32_t len)
     uint32_t slots_free;
     uint32_t slots_total;
     char *hub_name;
+    char *local_nick;
     struct sockaddr_in hub_addr;
 
     if (strncmp(buf, "$SR ", 4) != 0)
@@ -352,7 +454,10 @@ parse_search_response(char *buf, uint32_t len)
     token = strsep(&buf, " ");
     if (token == NULL)
     	return NULL; /* Invalid $SR message: Missing user. */
-    ui = hmap_get(hub_users, token);
+    local_nick = hub_to_main_string(token);
+    ui = hmap_get(hub_users, local_nick);
+    free(local_nick);
+
     if (ui == NULL)
     	return NULL; /* Invalid $SR message: Unknown user. */
 
@@ -374,32 +479,47 @@ parse_search_response(char *buf, uint32_t len)
     if (filename == buf)
     	return NULL; /* Invalid $SR message: Missing free slots. */
     *buf = '\0';
+
+	filename = hub_to_main_string(filename);
+    if (filename == NULL)
+        return NULL; /* Invalid $SR message: not convertable to local charset */
+
     buf++;
     token = strchr(filename, '\x05');
     if (token == NULL) {
-	filetype = DC_TYPE_DIR;
+        filetype = DC_TYPE_DIR;
     	filesize = 0;
     } else {
-    	filetype = DC_TYPE_REG;
-	*token = '\0';
-	if (!parse_uint64(token+1, &filesize))
-	    return NULL; /* Invalid $SR message: Invalid file size */
+        filetype = DC_TYPE_REG;
+        *token = '\0';
+        if (!parse_uint64(token+1, &filesize)) {
+            free(filename);
+            return NULL; /* Invalid $SR message: Invalid file size */
+        }
     }
 
     token = strsep(&buf, "/");
     assert(token != NULL);
-    if (!parse_uint32(token, &slots_free))
-    	return NULL; /* Invalid $SR message: Invalid free slots. */
+    if (!parse_uint32(token, &slots_free)) {
+        free(filename);
+        return NULL; /* Invalid $SR message: Invalid free slots. */
+    }
     token = strsep(&buf, "\x05");
-    if (token == NULL)
-    	return NULL; /* Invalid $SR message: Missing total slots. */
-    if (!parse_uint32(token, &slots_total))
-    	return NULL; /* Invalid $SR message: Invalid total slots. */
+    if (token == NULL) {
+        free(filename);
+        return NULL; /* Invalid $SR message: Missing total slots. */
+    }
+    if (!parse_uint32(token, &slots_total)) {
+        free(filename);
+        return NULL; /* Invalid $SR message: Invalid total slots. */
+    }
 
     hub_name = buf;
     token = strrchr(buf, '(');
-    if (token == NULL)
-    	return NULL; /* Invalid $SR message: Missing hub address. */
+    if (token == NULL) {
+        free(filename);
+        return NULL; /* Invalid $SR message: Missing hub address. */
+    }
     *token = '\0';
     buf = token+1;
     token = strchr(buf, ')');
@@ -412,20 +532,24 @@ parse_search_response(char *buf, uint32_t len)
 
     token = strchr(buf, ':');
     if (token != NULL) {
-    	*token = '\0';
-	if (!parse_uint16(token+1, &hub_addr.sin_port))
-    	    return NULL; /* Invalid $SR message: Invalid hub port. */
-	hub_addr.sin_port = htons(hub_addr.sin_port);
+        *token = '\0';
+        if (!parse_uint16(token+1, &hub_addr.sin_port)) {
+            free(filename);
+            return NULL; /* Invalid $SR message: Invalid hub port. */
+        }
+        hub_addr.sin_port = htons(hub_addr.sin_port);
     } else {
     	hub_addr.sin_port = htons(DC_HUB_TCP_PORT);
     }
-    if (!inet_aton(buf, &hub_addr.sin_addr))
-	return NULL; /* Invalid $SR message: Invalid hub address. */
+    if (!inet_aton(buf, &hub_addr.sin_addr)) {
+        free(filename);
+        return NULL; /* Invalid $SR message: Invalid hub address. */
+    }
 
     sr = xmalloc(sizeof(DCSearchResponse));
     sr->userinfo = ui;
     ui->refcount++;
-    sr->filename = xstrdup(filename);
+    sr->filename = filename;
     sr->filetype = filetype;
     sr->filesize = filesize;
     sr->slots_free = slots_free;
@@ -463,8 +587,9 @@ match_selection_against_response(DCSearchSelection *ss, DCSearchResponse *sr)
     } else {
     	if (ss->datatype == DC_SEARCH_FOLDERS)
     	    return false;
-	if (ss->datatype == DC_SEARCH_CHECKSUM) /* TTH not supported yet */
-	    return false;
+	
+//	if (ss->datatype == DC_SEARCH_CHECKSUM) /* TTH not supported yet */
+//	    return false;
     	if (sr->filesize < ss->size_min || sr->filesize > ss->size_max)
 	    return false;
 	if (!match_search_patterns(sr->filename, ss))
@@ -499,6 +624,7 @@ add_search_request(char *args)
     DCSearchRequest *sr = NULL;
     uint32_t c;
     time_t now;
+    char *hub_args;
 
     for (c = 0; args[c] != '\0'; c++) {
     	if (args[c] == '|' || args[c] == ' ')
@@ -509,48 +635,80 @@ add_search_request(char *args)
     sel.size_max = UINT64_MAX;
     sel.datatype = DC_SEARCH_ANY;
     if (!parse_search_strings(args, &sel)) {
-	warn(_("No pattern to match.\n"));
-	return false;
+        int i = 0;
+        if (sel.patterns != NULL) {
+            for (i = 0; i < sel.patterncount; i++) {
+                search_string_free(sel.patterns+i);
+            }
+            free(sel.patterns);
+        }
+
+        warn(_("No pattern to match.\n"));
+        return false;
     }
 
     for (c = 0; c < our_searches->cur; c++) {
     	sr = our_searches->buf[c];
     	if (compare_search_selection(&sel, &sr->selection) == 0)
-	    break;
+            break;
     }
 
     if (time(&now) == (time_t) -1) {
-	warn(_("Cannot get current time - %s\n"), errstr);
-	free(sel.patterns);
-	return false;
+        warn(_("Cannot get current time - %s\n"), errstr);
+        if (sel.patterns != NULL) {
+            int i = 0;
+            for (i = 0; i < sel.patterncount; i++) {
+                search_string_free(sel.patterns+i);
+            }
+            free(sel.patterns);
+        }
+        return false;
     }
 
     if (c < our_searches->cur) {
     	screen_putf(_("Reissuing search %d.\n"), c+1);
-    	free(sel.patterns);
-	sr->issue_time = now;
+        if (sel.patterns != NULL) {
+            int i = 0;
+            for (i = 0; i < sel.patterncount; i++) {
+                search_string_free(sel.patterns+i);
+            }
+            free(sel.patterns);
+        }
+        sr->issue_time = now;
     } else {
-    	screen_putf(_("Issuing new search with index %d.\n"), c+1);
-	sr = xmalloc(sizeof(DCSearchRequest));
-	sr->selection = sel;
-	sr->responses = ptrv_new();
-	sr->issue_time = now;
-	ptrv_append(our_searches, sr);
+        screen_putf(_("Issuing new search with index %d.\n"), c+1);
+        sr = xmalloc(sizeof(DCSearchRequest));
+        sr->selection = sel;
+        sr->responses = ptrv_new();
+        sr->issue_time = now;
+        ptrv_append(our_searches, sr);
     }
 
+
+    /* convert search string from local to hub charset */
+    hub_args = main_to_hub_string(args);
+
     if (is_active) {
-	hub_putf("$Search %s:%u F?F?0?1?%s|",
-    		inet_ntoa(local_addr.sin_addr), listen_port, args);
+        hub_putf("$Search %s:%u F?F?0?1?%s|", inet_ntoa(local_addr.sin_addr), listen_port, hub_args);
     } else {
-	hub_putf("$Search Hub:%s F?F?0?1?%s|", my_nick, args);
+        char *hub_my_nick;
+        hub_my_nick = main_to_hub_string(my_nick);
+        hub_putf("$Search Hub:%s F?F?0?1?%s|", hub_my_nick, hub_args);
+        free(hub_my_nick);
     }
 
+    free(hub_args);
+
     return true;
 }
 
 void
 free_search_request(DCSearchRequest *sr)
 {
+    int i = 0;
+    for (i = 0; sr->selection.patterns != NULL && i < sr->selection.patterncount; i++) {
+        search_string_free(sr->selection.patterns+i);
+    }
     free(sr->selection.patterns);
     ptrv_foreach(sr->responses, (PtrVForeachCallback) free_search_response);
     ptrv_free(sr->responses);
@@ -564,42 +722,42 @@ handle_search_result(char *buf, uint32_t len)
     time_t now;
 
     if (time(&now) == (time_t) -1) {
-	warn(_("Cannot get current time - %s\n"), errstr);
-	return;
+        warn(_("Cannot get current time - %s\n"), errstr);
+        return;
     }
 
     sr = parse_search_response(buf, len);
     if (sr == NULL) {
         warn(_("Unterminated or invalid $SR, discarding: %s\n"), quotearg_mem(buf, len));
-    	return;
+        return;
     }
 
     for (c = 0; c < our_searches->cur; c++) {
     	DCSearchRequest *sd = our_searches->buf[c];
 
-	if (sd->issue_time + SEARCH_TIME_THRESHOLD <= now)
-	    continue;
+        if (sd->issue_time + SEARCH_TIME_THRESHOLD <= now)
+            continue;
 
-   	if (match_selection_against_response(&sd->selection, sr)) {
-	    char *fmt;
-	    uint32_t d;
+        if (match_selection_against_response(&sd->selection, sr)) {
+            char *fmt;
+            uint32_t d;
 
-	    /* This is slow, but better than dupes... */
-	    for (d = 0; d < sd->responses->cur; d++) {
-		if (compare_search_response(sd->responses->buf[d], sr) == 0) {
-		    screen_putf(_("Result has been added earlier to search %d.\n"), c+1);
-		    break;
-		}
-	    }
-	    if (d >= sd->responses->cur) {
-		ptrv_append(sd->responses, sr);
-		fmt = ngettext("Added result to search %d (now %d result).\n",
-			       "Added result to search %d (now %d results).\n",
-			       sd->responses->cur);
-		flag_putf(DC_DF_SEARCH_RESULTS, fmt, c+1, sd->responses->cur);
-		sr->refcount++;
-	    }
-	}
+            /* This is slow, but better than dupes... */
+            for (d = 0; d < sd->responses->cur; d++) {
+                if (compare_search_response(sd->responses->buf[d], sr) == 0) {
+                    screen_putf(_("Result has been added earlier to search %d.\n"), c+1);
+                    break;
+                }
+            }
+            if (d >= sd->responses->cur) {
+                ptrv_append(sd->responses, sr);
+                fmt = ngettext("Added result to search %d (now %d result).\n",
+                               "Added result to search %d (now %d results).\n",
+                               sd->responses->cur);
+                flag_putf(DC_DF_SEARCH_RESULTS, fmt, c+1, sd->responses->cur);
+                sr->refcount++;
+            }
+        }
     }
 
     free_search_response(sr);
