@@ -95,31 +95,35 @@ ilog10(uint64_t c)
 
 /* XXX: try to get rid of screen_putf from this function. */
 int
-mkdirs_for_file(char *filename, bool deletedirs)
+mkdirs_for_file(char *filename)
 {
     char *t;
 
     for (t = filename; *t == '/'; t++);
     while ((t = strchr(t, '/')) != NULL) {
-	struct stat st;
+	    struct stat st;
+
+            *t = '\0';
+	    if (stat(filename, &st) < 0) {
+	        if (errno != ENOENT) {
+	    	    screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
+                return -1;
+            } else {
+                if (mkdir(filename, 0777) < 0) {
+                    screen_putf(_("%s: Cannot create directory - %s\n"), quotearg(filename), errstr); 
+                    return -1;
+                }
+                /*
+                else if (deletedirs) {
+	    	        if (ptrv_find(delete_dirs, filename, (comparison_fn_t) strcmp) < 0)
+	    		    ptrv_append(delete_dirs, xstrdup(filename));
+                }
+                */
+            }
+        }
 
-        *t = '\0';
-	if (stat(filename, &st) < 0) {
-	    if (errno != ENOENT) {
-	    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
-		return -1;
-	    } else {
-        	if (mkdir(filename, 0777) < 0) {
-    	    	    screen_putf(_("%s: Cannot create directory - %s\n"), quotearg(filename), errstr); 
-		    return -1;
-		} else if (deletedirs) {
-	    	    if (ptrv_find(delete_dirs, filename, (comparison_fn_t) strcmp) < 0)
-	    		ptrv_append(delete_dirs, xstrdup(filename));
-		}
-	    }
-	}
         *t = '/';
-	for (; *t == '/'; t++);
+	    for (; *t == '/'; t++);
     }
 
     return 0;
