@@ -25,12 +25,14 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#include <iconv.h>
 #include <langinfo.h>		/* POSIX (XSI) */
 #include "xalloc.h"		/* Gnulib */
 #include "quotearg.h"		/* Gnulib */
 #include "minmax.h"		/* Gnulib */
 #include "xvasprintf.h"		/* Gnulib */
 #include "strnlen.h"		/* Gnulib */
+#include "iconvme.h"
 #include "xstrndup.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
@@ -58,14 +60,15 @@
  */
 #define POSITIVE_BOOL_STRINGS _("1,on,yes,true")
 
-static void var_set_charset(DCVariable *var, int argc, char **argv);
+static void var_set_hub_charset(DCVariable *var, int argc, char **argv);
+static void var_set_fs_charset(DCVariable *var, int argc, char **argv);
+static void var_set_log_charset(DCVariable *var, int argc, char **argv);
 static void var_set_nick(DCVariable *var, int argc, char **argv);
 static void var_set_password(DCVariable *var, int argc, char **argv);
 static void var_set_description(DCVariable *var, int argc, char **argv);
 static void var_set_tag(DCVariable *var, int argc, char **argv);
 static void var_set_speed(DCVariable *var, int argc, char **argv);
 static void var_set_email(DCVariable *var, int argc, char **argv);
-static void var_set_share_dir(DCVariable *var, int argc, char **argv);
 static void var_set_download_dir(DCVariable *var, int argc, char **argv);
 static void var_set_listing_dir(DCVariable *var, int argc, char **argv);
 static void var_set_slots(DCVariable *var, int argc, char **argv);
@@ -73,6 +76,7 @@ static char *var_get_slots(DCVariable *var);
 static char *var_get_string(DCVariable *var);
 static char *var_get_bool(DCVariable *var);
 static void var_set_active(DCVariable *var, int argc, char **argv);
+static void var_set_auto_reconnect(DCVariable *var, int argc, char **argv);
 static char *var_get_listen_addr(DCVariable *var);
 static void var_set_listen_addr(DCVariable *var, int argc, char **argv);
 static char *var_get_listen_port(DCVariable *var);
@@ -80,6 +84,9 @@ static void var_set_listen_port(DCVariable *var, int argc, char **argv);
 static char *var_get_display_flags(DCVariable *var);
 static void var_set_display_flags(DCVariable *var, int argc, char **argv);
 static void var_set_log_file(DCVariable *var, int argc, char **argv);
+static char *var_get_time(DCVariable *var);
+static void var_set_filelist_refresh_interval(DCVariable *var, int argc, char **argv);
+
 static void speed_completion_generator(DCCompletionInfo *ci);
 static void bool_completion_generator(DCCompletionInfo *ci);
 static void variable_completion_generator(DCCompletionInfo *ci);
@@ -92,6 +99,9 @@ typedef struct {
 } DCDisplayFlagDetails;
 
 /* settings/variables */
+uint32_t minislot_count = 3;
+uint32_t minislot_size = (1 << 16);
+int used_mini_slots = 0;
 int used_ul_slots = 0;
 int used_dl_slots = 0;
 uint32_t my_ul_slots;
@@ -105,6 +115,7 @@ char *download_dir;
 char *listing_dir;
 char *my_password;
 bool is_active;
+bool auto_reconnect = 0;
 uint64_t my_share_size = 0;
 uint32_t display_flags = ~(DC_DF_DEBUG); /* All flags except debug set */
 uint32_t log_flags = ~(DC_DF_DEBUG);
@@ -112,13 +123,13 @@ struct in_addr force_listen_addr = { INADDR_NONE };
 
 /* This list must be sorted according to strcmp. */
 DCDisplayFlagDetails display_flag_details[] =  {
-    { DC_DF_CONNECTIONS, 	"connections" },
-    { DC_DF_DEBUG, 		"debug" },
-    { DC_DF_DOWNLOAD, 		"download" },
-    { DC_DF_JOIN_PART, 		"joinpart" },
-    { DC_DF_PUBLIC_CHAT, 	"publicchat" },
-    { DC_DF_SEARCH_RESULTS, 	"searchresults" },
-    { DC_DF_UPLOAD, 		"upload" },
+    { DC_DF_CONNECTIONS,    "connections" },
+    { DC_DF_DEBUG,          "debug" },
+    { DC_DF_DOWNLOAD,       "download" },
+    { DC_DF_JOIN_PART,      "joinpart" },
+    { DC_DF_PUBLIC_CHAT,    "publicchat" },
+    { DC_DF_SEARCH_RESULTS, "searchresults" },
+    { DC_DF_UPLOAD,         "upload" },
 };
 static const int display_flag_count = sizeof(display_flag_details)/sizeof(*display_flag_details);
 
@@ -146,11 +157,11 @@ static DCVariable variables[] = {
       "Enable if listening for remote connections"
     },
     {
-      "charset",
-      var_get_string, var_set_charset, &hub_charset,
-      charset_completion_generator,
+      "auto_reconnect",
+      var_get_bool, var_set_auto_reconnect, &auto_reconnect,
+      bool_completion_generator,
       NULL,
-      "Character set used for chat on the hub"
+      "Enable automatic reconnect to the last connected hub"
     },
     {
       "description",
@@ -180,6 +191,27 @@ static DCVariable variables[] = {
       NULL,
       "The e-mail visible to other users of the hub"
     },
+    {
+      "filelist_refresh_interval",
+      var_get_time, var_set_filelist_refresh_interval, &filelist_refresh_timeout,
+      NULL,
+      NULL,
+      "Local filelist refresh interval (in seconds)"
+    },
+    {
+      "filesystem_charset",
+      var_get_string, var_set_fs_charset, &fs_charset,
+      charset_completion_generator,
+      NULL,
+      "Local filesystem charset (if it differs from local charset)"
+    },
+    {
+      "hub_charset",
+      var_get_string, var_set_hub_charset, &hub_charset,
+      charset_completion_generator,
+      NULL,
+      "Character set used for chat on the hub"
+    },
     {
       "listenaddr",
       var_get_listen_addr, var_set_listen_addr, &force_listen_addr,
@@ -208,6 +240,13 @@ static DCVariable variables[] = {
       NULL,
       "Types of messages to log (if logfile set)"
     },
+    {
+      "log_charset",
+      var_get_string, var_set_log_charset, &log_charset,
+      charset_completion_generator,
+      NULL,
+      "Log charset (if it differs from local charset)"
+    },
     {
       "logfile",
       var_get_string, var_set_log_file, &log_filename,
@@ -229,6 +268,7 @@ static DCVariable variables[] = {
       NULL,
       "The optional password to pass to the hub."
     },
+/*
     {
       "sharedir",
       var_get_string, var_set_share_dir, &share_dir,
@@ -236,6 +276,7 @@ static DCVariable variables[] = {
       NULL,
       "Directory containing files to share (or a single file to share)"
     },
+*/
     {
       "slots",
       var_get_slots, var_set_slots, &my_ul_slots,
@@ -278,43 +319,42 @@ charset_completion_generator(DCCompletionInfo *ci)
     exec('iconv', 'iconv', '--list');*/
 }
 
-static void
-var_set_charset(DCVariable *var, int argc, char **argv)
-{
-    iconv_t new_from_iconv = (iconv_t) -1;
-    iconv_t new_to_iconv = (iconv_t) -1;
 
+static void
+var_set_hub_charset(DCVariable *var, int argc, char **argv)
+{
     if (argc > 2) {
         warn(_("too many arguments\n"));
         return;
     }
-    if (argv[1][0] != '\0') {
-	char *system_charset;
+    set_hub_charset(argv[1]);
 
-  	system_charset = nl_langinfo(CODESET);
-  	if (system_charset == NULL || *system_charset == '\0') {
-    	    warn(_("cannot get current character set"));
-	    return;
-	}
-        new_to_iconv = iconv_open(argv[1], system_charset);
-        if (new_to_iconv == (iconv_t) -1) {
-            warn(_("cannot create character set convertor - %s\n"), errstr);
-            return;
-        }
-        new_from_iconv = iconv_open(system_charset, argv[1]);
-        if (new_from_iconv == (iconv_t) -1) {
-            warn(_("cannot create character set convertor - %s\n"), errstr);
-            return;
-        }
+    /* rebuild filelist (filenames should be in hub charset ) */
+    update_request_set_hub_charset(argv[1]);
+    if (hub_state == DC_HUB_LOGGED_IN)
+        hub_reconnect();
+}
+
+static void
+var_set_fs_charset(DCVariable *var, int argc, char **argv)
+{
+    if (argc > 2) {
+        warn(_("too many arguments\n"));
+        return;
     }
-    if (to_hub_iconv != (iconv_t) -1 && iconv_close(to_hub_iconv) != 0)
-        warn(_("cannot close character convertor - %s\n"), errstr);
-    if (from_hub_iconv != (iconv_t) -1 && iconv_close(from_hub_iconv) != 0)
-        warn(_("cannot close character convertor - %s\n"), errstr);
-    to_hub_iconv = new_to_iconv;
-    from_hub_iconv = new_from_iconv;
-    free(hub_charset);
-    hub_charset = argv[1][0] == '\0' ? NULL : xstrdup(argv[1]);
+    set_fs_charset(argv[1]);
+    /* rebuild filelist (read filenames in new charset */
+    update_request_set_fs_charset(argv[1]);
+}
+
+static void
+var_set_log_charset(DCVariable *var, int argc, char **argv)
+{
+    if (argc > 2) {
+        warn(_("too many arguments\n"));
+        return;
+    }
+    set_log_charset(argv[1]);
 }
 
 static void
@@ -394,26 +434,6 @@ var_set_speed(DCVariable *var, int argc, char **argv)
     }
 }
 
-static void
-var_set_share_dir(DCVariable *var, int argc, char **argv)
-{
-    struct stat st;
-
-    if (argc > 2) {
-	warn(_("too many arguments\n"));
-	return;
-    }
-    if (stat(argv[1], &st) < 0) {
-    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(argv[1]), errstr);
-	return;
-    }
-    if (!S_ISDIR(st.st_mode)) {
-    	screen_putf(_("%s: Not a directory\n"), quotearg(argv[1]));
-	return;
-    }
-    set_share_dir(argv[1]); /* Ignore errors */
-}
-
 static void
 var_set_download_dir(DCVariable *var, int argc, char **argv)
 {
@@ -454,7 +474,7 @@ var_set_listing_dir(DCVariable *var, int argc, char **argv)
     }
     free(listing_dir);
     listing_dir = xstrdup(argv[1]);
-    set_share_dir(share_dir);
+    update_request_set_listing_dir(listing_dir);
 }
 
 static void
@@ -466,6 +486,9 @@ var_set_slots(DCVariable *var, int argc, char **argv)
     }
     if (!parse_uint32(argv[1], &my_ul_slots))
     	screen_putf(_("Invalid slot number `%s'\n"), quotearg(argv[1]));
+
+    if (hub_state >= DC_HUB_LOGGED_IN)
+        send_my_info();
 }
 
 static char *
@@ -554,11 +577,27 @@ var_set_active(DCVariable *var, int argc, char **argv)
     	screen_putf(_("Active setting not changed.\n"));/*return false;*/
 }
 
+static void
+var_set_auto_reconnect(DCVariable *var, int argc, char **argv)
+{
+    bool state;
+    
+    if (argc > 2) {
+	warn(_("too many arguments\n"));
+	return;
+    }
+    if (!parse_bool(argv[1], &state)) {
+	screen_putf(_("Specify value as `0', `no', `off', `1', `yes', or `on'.\n"));
+	return;/*false*/
+    }
+    auto_reconnect = state;
+}
+
 static char *
 var_get_listen_addr(DCVariable *var)
 {
     if (force_listen_addr.s_addr == INADDR_NONE)
-	return NULL;
+        return NULL;
     return xstrdup(in_addr_str(force_listen_addr));
 }
 
@@ -568,13 +607,13 @@ var_set_listen_addr(DCVariable *var, int argc, char **argv)
     struct sockaddr_in addr;
 
     if (argc > 2) {
-	warn(_("too many arguments\n"));
-	return;
+        warn(_("too many arguments\n"));
+        return;
     }
     if (argv[1][0] == '\0') {
-	force_listen_addr.s_addr = INADDR_NONE;
-	screen_putf(_("Removing listening address.\n"));
-	return;
+        force_listen_addr.s_addr = INADDR_NONE;
+        screen_putf(_("Removing listening address.\n"));
+        return;
     }
 
     if (!inet_aton(argv[1], &addr.sin_addr)) {
@@ -745,6 +784,34 @@ var_set_password(DCVariable *var, int argc, char **argv)
     }
 }
 
+static char*
+var_get_time(DCVariable *var)
+{
+    time_t value = *((time_t *) var->value);
+    return xasprintf("%ld", value);
+}
+
+static void
+var_set_filelist_refresh_interval(DCVariable *var, int argc, char **argv)
+{
+    unsigned int interval;
+    if (argc > 2) {
+        warn(_("too many arguments\n"));
+        return;
+    }
+    if (argv[1][0] == '\0') {
+        interval = 0;
+    } else {
+        if (!parse_uint32(argv[1], &interval)) {
+            screen_putf(_("Invalid value `%s' for interval.\n"), quotearg(argv[1]));
+            return;
+        }
+    }
+    filelist_refresh_timeout = interval;
+
+    update_request_set_filelist_refresh_timeout(filelist_refresh_timeout);
+}
+
 /* The following completers could be improved by doing a modified
  * binsearch for the first match, then stopping immediately when a
  * non-match is found.
